import pytest
from cognite.client.data_classes.capabilities import (
    AssetsAcl,
    Capability,
    DataModelInstancesAcl,
    DataModelsAcl,
    EventsAcl,
    ExtractionPipelinesAcl,
    FilesAcl,
    LabelsAcl,
    ProjectCapability,
    ProjectCapabilityList,
    RelationshipsAcl,
    SequencesAcl,
    ThreeDAcl,
    TimeSeriesAcl,
    TransformationsAcl,
    WorkflowOrchestrationAcl,
)
from cognite.client.data_classes.iam import TokenInspection

from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.exceptions import AuthorizationError
from cognite_toolkit._cdf_tk.utils.validate_access import ValidateAccess


class TestValidateAccess:
    @pytest.mark.parametrize(
        "capabilities, spaces, expected_error",
        [
            pytest.param(
                [],
                None,
                "You have no permission to read data models. This is required to test the operation.",
                id="No capabilities",
            ),
            pytest.param(
                [DataModelsAcl([DataModelsAcl.Action.Read], DataModelsAcl.Scope.SpaceID(["space1"]))],
                {"space2"},
                "You have no permission to read the 'space2' space(s). This is required to test the operation.",
                id="Space mismatch",
            ),
        ],
    )
    def test_model_access_raise(
        self, capabilities: list[Capability], spaces: set[str] | None, expected_error: str
    ) -> None:
        inspection = self._create_inspection_obj(capabilities)

        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = inspection
            validator = ValidateAccess(client, "test the operation")
            with pytest.raises(AuthorizationError) as exc:
                validator.data_model(["read"], spaces=spaces)
            assert str(exc.value) == expected_error

    @pytest.mark.parametrize(
        "capabilities, spaces, expected_result",
        [
            pytest.param(
                [DataModelsAcl([DataModelsAcl.Action.Read], DataModelsAcl.Scope.SpaceID(["space1"]))],
                {"space1"},
                None,
                id="Space match",
            ),
            pytest.param(
                [DataModelsAcl([DataModelsAcl.Action.Read], DataModelsAcl.Scope.All())], None, None, id="All scope"
            ),
            pytest.param(
                [DataModelsAcl([DataModelsAcl.Action.Read], DataModelsAcl.Scope.SpaceID(["space1", "space2"]))],
                None,
                ["space1", "space2"],
                id="Limited list of spaces",
            ),
        ],
    )
    def test_model_access(
        self, capabilities: list[Capability], spaces: set[str] | None, expected_result: list[str] | None
    ) -> None:
        inspection = self._create_inspection_obj(capabilities)

        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = inspection
            validator = ValidateAccess(client, "test the operation")
            result = validator.data_model(["read"], spaces=spaces)
            assert result == expected_result

    @pytest.mark.parametrize(
        "capabilities, spaces, expected_error",
        [
            pytest.param(
                [],
                None,
                "You have no permission to read instances. This is required to test the operation.",
                id="No capabilities",
            ),
            pytest.param(
                [
                    DataModelInstancesAcl(
                        [DataModelInstancesAcl.Action.Read], DataModelInstancesAcl.Scope.SpaceID(["space1"])
                    )
                ],
                {"space2"},
                "You have no permission to read instances in the 'space2' space(s). This is required to test the operation instances.",
                id="Space mismatch",
            ),
        ],
    )
    def test_instances_access_raise(
        self, capabilities: list[Capability], spaces: set[str] | None, expected_error: str
    ) -> None:
        inspection = self._create_inspection_obj(capabilities)

        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = inspection
            validator = ValidateAccess(client, "test the operation")
            with pytest.raises(AuthorizationError) as exc:
                validator.instances(["read"], spaces=spaces)
            assert str(exc.value) == expected_error

    @pytest.mark.parametrize(
        "capabilities, spaces, expected_result",
        [
            pytest.param(
                [
                    DataModelInstancesAcl(
                        [DataModelInstancesAcl.Action.Read], DataModelInstancesAcl.Scope.SpaceID(["space1"])
                    )
                ],
                {"space1"},
                None,
                id="Space match",
            ),
            pytest.param(
                [DataModelInstancesAcl([DataModelInstancesAcl.Action.Read], DataModelInstancesAcl.Scope.All())],
                None,
                None,
                id="All scope",
            ),
            pytest.param(
                [
                    DataModelInstancesAcl(
                        [DataModelInstancesAcl.Action.Read], DataModelInstancesAcl.Scope.SpaceID(["space1", "space2"])
                    )
                ],
                None,
                ["space1", "space2"],
                id="Limited list of spaces",
            ),
        ],
    )
    def test_instances_access(
        self, capabilities: list[Capability], spaces: set[str] | None, expected_result: list[str] | None
    ) -> None:
        inspection = self._create_inspection_obj(capabilities)

        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = inspection
            validator = ValidateAccess(client, "test the operation")
            result = validator.instances(["read"], spaces=spaces)
            assert result == expected_result

    @pytest.mark.parametrize(
        "capabilities, dataset_ids, expected_error",
        [
            pytest.param(
                [],
                None,
                "You have no permission to read time series. This is required to test the operation.",
                id="No capabilities",
            ),
            pytest.param(
                [TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.DataSet([1]))],
                {2},
                "You have no permission to read time series in dataset 2. This is required to test the operation.",
                id="Dataset mismatch",
            ),
            pytest.param(
                [TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.AssetRootID([10]))],
                {1},
                "You have no permission to read time series in dataset 1. This is required to test the operation.",
                id="Access by other scope type",
            ),
        ],
    )
    def test_timeseries_access_raise(
        self, capabilities: list[Capability], dataset_ids: set[int] | None, expected_error: str
    ) -> None:
        inspection = self._create_inspection_obj(capabilities)

        def external_id_lookup(ids: list[int]) -> list[str]:
            return [str(id_) for id_ in ids]

        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = inspection
            client.lookup.data_sets.external_id.side_effect = external_id_lookup
            client.lookup.assets.external_id.side_effect = external_id_lookup
            client.lookup.time_series.external_id.side_effect = external_id_lookup

            validator = ValidateAccess(client, "test the operation")
            with pytest.raises(AuthorizationError) as exc:
                validator.timeseries(["read"], dataset_ids=dataset_ids)
            assert str(exc.value) == expected_error

    @pytest.mark.parametrize(
        "capabilities, dataset_ids, expected_result",
        [
            pytest.param(
                [TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.DataSet([1, 2]))],
                {1},
                None,
                id="Dataset match",
            ),
            pytest.param(
                [TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.All())],
                None,
                None,
                id="All scope",
            ),
            pytest.param(
                [TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.DataSet([1, 2]))],
                None,
                {"dataset": ["1", "2"]},
                id="Limited list of datasets",
            ),
            pytest.param(
                [
                    TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.AssetRootID([10, 20])),
                    TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.DataSet([15, 25])),
                    TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.ID([100, 200])),
                ],
                None,
                {
                    "asset root": ["10", "20"],
                    "dataset": ["15", "25"],
                    "time series": ["100", "200"],
                },
                id="Multiple scopes with IDs",
            ),
        ],
    )
    def test_timeseries_access(
        self, capabilities: list[Capability], dataset_ids: set[int] | None, expected_result: dict | None
    ) -> None:
        inspection = self._create_inspection_obj(capabilities)

        def external_id_lookup(ids: list[int]) -> list[str]:
            return [str(id_) for id_ in ids]

        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = inspection
            client.lookup.data_sets.external_id.side_effect = external_id_lookup
            client.lookup.assets.external_id.side_effect = external_id_lookup
            client.lookup.time_series.external_id.side_effect = external_id_lookup

            validator = ValidateAccess(client, "test the operation")
            result = validator.timeseries(["read"], dataset_ids=dataset_ids)
            assert result == expected_result

    @pytest.mark.parametrize(
        "capabilities, dataset_ids, expected_error",
        [
            pytest.param(
                [],
                None,
                "You have no permission to read files. This is required to test the operation.",
                id="No capabilities",
            ),
            pytest.param(
                [FilesAcl([FilesAcl.Action.Read], FilesAcl.Scope.DataSet([1]))],
                {2},
                "You have no permission to read files in dataset 2. This is required to test the operation.",
                id="Dataset mismatch",
            ),
        ],
    )
    def test_files_access_raise(
        self, capabilities: list[Capability], dataset_ids: set[int] | None, expected_error: str
    ) -> None:
        inspection = self._create_inspection_obj(capabilities)

        def external_id_lookup(ids: list[int]) -> list[str]:
            return [str(id_) for id_ in ids]

        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = inspection
            client.lookup.data_sets.external_id.side_effect = external_id_lookup
            validator = ValidateAccess(client, "test the operation")
            with pytest.raises(AuthorizationError) as exc:
                validator.files(["read"], dataset_ids=dataset_ids)
            assert str(exc.value) == expected_error

    @pytest.mark.parametrize(
        "capabilities, dataset_ids, expected_result",
        [
            pytest.param(
                [FilesAcl([FilesAcl.Action.Read], FilesAcl.Scope.DataSet([1, 2]))],
                {1},
                None,
                id="Dataset match",
            ),
            pytest.param(
                [FilesAcl([FilesAcl.Action.Read], FilesAcl.Scope.All())],
                None,
                None,
                id="All scope",
            ),
            pytest.param(
                [FilesAcl([FilesAcl.Action.Read], FilesAcl.Scope.DataSet([1, 2]))],
                None,
                {"dataset": ["1", "2"]},
                id="Limited list of datasets",
            ),
        ],
    )
    def test_files_access(
        self, capabilities: list[Capability], dataset_ids: set[int] | None, expected_result: dict | None
    ) -> None:
        inspection = self._create_inspection_obj(capabilities)

        def external_id_lookup(ids: list[int]) -> list[str]:
            return [str(id_) for id_ in ids]

        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = inspection
            client.lookup.data_sets.external_id.side_effect = external_id_lookup
            validator = ValidateAccess(client, "test the operation")
            result = validator.files(["read"], dataset_ids=dataset_ids)
            assert result == expected_result

    @pytest.mark.parametrize(
        "capabilities, dataset_ids, expected_error",
        [
            pytest.param(
                [],
                None,
                "You have no permission to read assets. This is required to test the operation.",
                id="No capabilities",
            ),
            pytest.param(
                [AssetsAcl([AssetsAcl.Action.Read], AssetsAcl.Scope.DataSet([1]))],
                {1, 2},
                "You have no permission to read assets in dataset(s) 2. This is required to test the operation.",
                id="Missing dataset",
            ),
        ],
    )
    def test_assets_access_raise(
        self, capabilities: list[Capability], dataset_ids: set[int] | None, expected_error: str
    ) -> None:
        inspection = self._create_inspection_obj(capabilities)

        def external_id_lookup(ids: list[int]) -> list[str]:
            return [str(id_) for id_ in ids]

        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = inspection
            client.lookup.data_sets.external_id.side_effect = external_id_lookup
            validator = ValidateAccess(client, "test the operation")
            with pytest.raises(AuthorizationError) as exc:
                validator.assets(["read"], dataset_ids=dataset_ids)
            assert str(exc.value) == expected_error

    @pytest.mark.parametrize(
        "capabilities, dataset_ids, expected_result",
        [
            pytest.param(
                [AssetsAcl([AssetsAcl.Action.Read], AssetsAcl.Scope.DataSet([1, 2]))],
                {1},
                None,
                id="Dataset match",
            ),
            pytest.param(
                [AssetsAcl([AssetsAcl.Action.Read], AssetsAcl.Scope.All())],
                None,
                None,
                id="All scope",
            ),
            pytest.param(
                [AssetsAcl([AssetsAcl.Action.Read], AssetsAcl.Scope.DataSet([1, 2]))],
                None,
                {"dataset": ["1", "2"]},
                id="Limited list of datasets",
            ),
        ],
    )
    def test_assets_access(
        self, capabilities: list[Capability], dataset_ids: set[int] | None, expected_result: dict | None
    ) -> None:
        inspection = self._create_inspection_obj(capabilities)

        def external_id_lookup(ids: list[int]) -> list[str]:
            return [str(id_) for id_ in ids]

        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = inspection
            client.lookup.data_sets.external_id.side_effect = external_id_lookup
            validator = ValidateAccess(client, "test the operation")
            result = validator.assets(["read"], dataset_ids=dataset_ids)
            assert result == expected_result

    @pytest.mark.parametrize(
        "capabilities, dataset_ids, expected_result",
        [
            pytest.param(
                [
                    AssetsAcl([AssetsAcl.Action.Read], AssetsAcl.Scope.DataSet([1, 2])),
                    FilesAcl([FilesAcl.Action.Read], FilesAcl.Scope.DataSet([1, 2])),
                    TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.DataSet([1])),
                    EventsAcl([EventsAcl.Action.Read], EventsAcl.Scope.DataSet([1])),
                    RelationshipsAcl([RelationshipsAcl.Action.Read], RelationshipsAcl.Scope.DataSet([1, 2])),
                    SequencesAcl([SequencesAcl.Action.Read], SequencesAcl.Scope.DataSet([1, 2])),
                    ThreeDAcl([ThreeDAcl.Action.Read], ThreeDAcl.Scope.DataSet([1])),
                    LabelsAcl([LabelsAcl.Action.Read], LabelsAcl.Scope.DataSet([1])),
                ],
                {1},
                None,
                id="Dataset match",
            ),
            pytest.param(
                [
                    AssetsAcl([AssetsAcl.Action.Read], AssetsAcl.Scope.All()),
                    FilesAcl([FilesAcl.Action.Read], FilesAcl.Scope.All()),
                    TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.All()),
                    EventsAcl([EventsAcl.Action.Read], EventsAcl.Scope.All()),
                    RelationshipsAcl([RelationshipsAcl.Action.Read], RelationshipsAcl.Scope.All()),
                    SequencesAcl([SequencesAcl.Action.Read], SequencesAcl.Scope.All()),
                    ThreeDAcl([ThreeDAcl.Action.Read], ThreeDAcl.Scope.All()),
                    LabelsAcl([LabelsAcl.Action.Read], LabelsAcl.Scope.All()),
                ],
                None,
                None,
                id="All scope",
            ),
            pytest.param(
                [
                    AssetsAcl([AssetsAcl.Action.Read], AssetsAcl.Scope.DataSet([1, 2])),
                    FilesAcl([FilesAcl.Action.Read], FilesAcl.Scope.DataSet([1, 3])),
                    TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.DataSet([1, 4])),
                    EventsAcl([EventsAcl.Action.Read], EventsAcl.Scope.DataSet([1, 2])),
                    RelationshipsAcl([RelationshipsAcl.Action.Read], RelationshipsAcl.Scope.DataSet([1, 2])),
                    SequencesAcl([SequencesAcl.Action.Read], SequencesAcl.Scope.DataSet([1, 9])),
                    ThreeDAcl([ThreeDAcl.Action.Read], ThreeDAcl.Scope.DataSet([1, 3])),
                    LabelsAcl([LabelsAcl.Action.Read], LabelsAcl.Scope.DataSet([1, 3])),
                ],
                None,
                {
                    "3D models": [1, 3],
                    "assets": [1, 2],
                    "events": [1, 2],
                    "files": [1, 3],
                    "labels": [1, 3],
                    "relationships": [1, 2],
                    "sequences": [1, 9],
                    "time series": [1, 4],
                },
                id="Limited list of datasets",
            ),
        ],
    )
    def test_dataset_data_access(
        self, capabilities: list[Capability], dataset_ids: set[int] | None, expected_result: dict | None
    ) -> None:
        inspection = self._create_inspection_obj(capabilities)
        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = inspection
            validator = ValidateAccess(client, "test the operation")
            result = validator.dataset_data(["read"], dataset_ids=dataset_ids)
            assert result == expected_result

    @pytest.mark.parametrize(
        "capabilities, expected_error",
        [
            pytest.param(
                [],
                (
                    "You have no permission to read 3D models, assets, events, files, labels, "
                    "relationships, sequences and time series on datasets 1 and 2. This is "
                    "required to test the operation."
                ),
                id="No capabilities",
            ),
            pytest.param(
                [
                    AssetsAcl([AssetsAcl.Action.Read], AssetsAcl.Scope.All()),
                    FilesAcl([FilesAcl.Action.Read], FilesAcl.Scope.All()),
                ],
                (
                    "You have no permission to read 3D models, events, labels, relationships, "
                    "sequences and time series on datasets 1 and 2. This is required to test the "
                    "operation."
                ),
                id="Partial capabilities",
            ),
        ],
    )
    def test_dataset_no_data_access(self, capabilities: list[Capability], expected_error: str) -> None:
        inspection = self._create_inspection_obj(capabilities)

        def external_id_lookup(ids: list[int]) -> list[str]:
            return [str(id_) for id_ in ids]

        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = inspection
            client.lookup.data_sets.external_id.side_effect = external_id_lookup
            validator = ValidateAccess(client, "test the operation")
            with pytest.raises(AuthorizationError) as exc:
                validator.dataset_data(["read"], dataset_ids={1, 2})
            assert str(exc.value) == expected_error

    @pytest.mark.parametrize(
        "capabilities, dataset_ids, expected_result",
        [
            pytest.param(
                [
                    TransformationsAcl([TransformationsAcl.Action.Read], TransformationsAcl.Scope.DataSet([1, 2])),
                    WorkflowOrchestrationAcl(
                        [WorkflowOrchestrationAcl.Action.Read], WorkflowOrchestrationAcl.Scope.DataSet([1, 2])
                    ),
                    ExtractionPipelinesAcl(
                        [ExtractionPipelinesAcl.Action.Read], ExtractionPipelinesAcl.Scope.DataSet([1])
                    ),
                ],
                {1},
                None,
                id="Dataset match",
            ),
            pytest.param(
                [
                    TransformationsAcl([TransformationsAcl.Action.Read], TransformationsAcl.Scope.All()),
                    WorkflowOrchestrationAcl(
                        [WorkflowOrchestrationAcl.Action.Read], WorkflowOrchestrationAcl.Scope.All()
                    ),
                    ExtractionPipelinesAcl([ExtractionPipelinesAcl.Action.Read], ExtractionPipelinesAcl.Scope.All()),
                ],
                None,
                None,
                id="All scope",
            ),
            pytest.param(
                [
                    TransformationsAcl([TransformationsAcl.Action.Read], TransformationsAcl.Scope.DataSet([1, 2])),
                    WorkflowOrchestrationAcl(
                        [WorkflowOrchestrationAcl.Action.Read], WorkflowOrchestrationAcl.Scope.DataSet([1, 3])
                    ),
                    ExtractionPipelinesAcl(
                        [ExtractionPipelinesAcl.Action.Read], ExtractionPipelinesAcl.Scope.DataSet([1, 4])
                    ),
                ],
                None,
                {
                    "transformations": [1, 2],
                    "workflows": [1, 3],
                    "extraction pipelines": [1, 4],
                },
                id="Limited list of datasets",
            ),
        ],
    )
    def test_dataset_configuration_access(
        self, capabilities: list[Capability], dataset_ids: set[int] | None, expected_result: dict | None
    ) -> None:
        inspection = self._create_inspection_obj(capabilities)
        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = inspection
            validator = ValidateAccess(client, "test the operation")
            result = validator.dataset_configurations(["read"], dataset_ids=dataset_ids)
            assert result == expected_result

    @pytest.mark.parametrize(
        "capabilities, expected_error",
        [
            pytest.param(
                [],
                (
                    "You have no permission to read extraction pipelines, transformations "
                    "and workflows on datasets 1 and 2. This is required to test the operation."
                ),
                id="No capabilities",
            ),
            pytest.param(
                [
                    TransformationsAcl([TransformationsAcl.Action.Read], TransformationsAcl.Scope.All()),
                ],
                (
                    "You have no permission to read extraction pipelines and workflows on datasets 1 and 2. "
                    "This is required to test the operation."
                ),
                id="Partial capabilities",
            ),
            pytest.param(
                [
                    TransformationsAcl([TransformationsAcl.Action.Read], TransformationsAcl.Scope.All()),
                    WorkflowOrchestrationAcl(
                        [WorkflowOrchestrationAcl.Action.Read], WorkflowOrchestrationAcl.Scope.All()
                    ),
                    ExtractionPipelinesAcl(
                        [ExtractionPipelinesAcl.Action.Read], ExtractionPipelinesAcl.Scope.DataSet([1])
                    ),
                ],
                (
                    "You have no permission to read extraction pipelines on datasets 1 and 2. "
                    "This is required to test the operation."
                ),
                id="Missing dataset access",
            ),
        ],
    )
    def test_dataset_no_configuration_access(self, capabilities: list[Capability], expected_error: str) -> None:
        inspection = self._create_inspection_obj(capabilities)

        def external_id_lookup(ids: list[int]) -> list[str]:
            return [str(id_) for id_ in ids]

        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = inspection
            client.lookup.data_sets.external_id.side_effect = external_id_lookup
            validator = ValidateAccess(client, "test the operation")
            with pytest.raises(AuthorizationError) as exc:
                validator.dataset_configurations(["read"], dataset_ids={1, 2})
            assert str(exc.value) == expected_error

    @staticmethod
    def _create_inspection_obj(capabilities: list[Capability]) -> TokenInspection:
        inspection = TokenInspection(
            "123",
            [],
            ProjectCapabilityList(
                [
                    ProjectCapability(
                        capability=capability,
                        project_scope=ProjectCapability.Scope.All(),
                    )
                    for capability in capabilities
                ]
            ),
        )
        return inspection



import shutil
import subprocess
from contextlib import suppress
from pathlib import Path


def use_poetry() -> bool:
    with suppress(Exception):
        return shutil.which("poetry") is not None
    return False


def use_git() -> bool:
    with suppress(Exception):
        return shutil.which("git") is not None
    return False


def has_initiated_repo() -> bool:
    with suppress(Exception):
        result = subprocess.run("git rev-parse --is-inside-work-tree".split(), stdout=subprocess.PIPE)
        return result.returncode == 0
    return False


def has_uncommitted_changes() -> bool:
    with suppress(Exception):
        result = subprocess.run("git diff --quiet".split(), stdout=subprocess.PIPE)
        return result.returncode != 0
    return False


def git_root() -> Path | None:
    with suppress(Exception):
        result = subprocess.run("git rev-parse --show-toplevel".split(), stdout=subprocess.PIPE)
        return Path(result.stdout.decode(encoding="utf-8").strip())
    return None



from __future__ import annotations

import os
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
import yaml
from _pytest.monkeypatch import MonkeyPatch
from cognite.client.data_classes.data_modeling import DataModelId, Space

from cognite_toolkit._cdf_tk.commands.build_cmd import BuildCommand
from cognite_toolkit._cdf_tk.cruds import TransformationCRUD
from cognite_toolkit._cdf_tk.data_classes import BuildConfigYAML, BuildVariables, Environment, Packages
from cognite_toolkit._cdf_tk.data_classes._module_directories import ModuleDirectories
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitMissingModuleError,
)
from cognite_toolkit._cdf_tk.feature_flags import Flags
from cognite_toolkit._cdf_tk.hints import ModuleDefinition
from cognite_toolkit._cdf_tk.tk_warnings import LowSeverityWarning
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests import data
from tests.test_unit.approval_client import ApprovalToolkitClient


@pytest.fixture(scope="session")
def dummy_environment() -> Environment:
    return Environment(
        name="dev",
        project="my_project",
        validation_type="dev",
        selected=["none"],
    )


class TestBuildCommand:
    def test_module_not_found_error(self, tmp_path: Path) -> None:
        with pytest.raises(ToolkitMissingModuleError):
            BuildCommand(print_warning=False).execute(
                verbose=False,
                build_dir=tmp_path,
                organization_dir=data.PROJECT_WITH_BAD_MODULES,
                selected=None,
                build_env_name="no_module",
                no_clean=False,
            )

    def test_module_with_non_resource_directories(self, tmp_path: Path) -> None:
        cmd = BuildCommand(print_warning=False)
        cmd.execute(
            verbose=False,
            build_dir=tmp_path,
            organization_dir=data.PROJECT_WITH_BAD_MODULES,
            selected=None,
            build_env_name="ill_module",
            no_clean=False,
        )

        assert len(cmd.warning_list) >= 1
        assert (
            LowSeverityWarning(
                f"Module 'ill_made_module' has non-resource directories: ['spaces']. {ModuleDefinition.short()}"
            )
            in cmd.warning_list
        )

    @pytest.mark.skipif(not Flags.GRAPHQL.is_enabled(), reason="GraphQL schema files will give warnings")
    def test_custom_project_no_warnings(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        cmd = BuildCommand(print_warning=False)
        monkeypatch.setenv("CDF_PROJECT", "some-project")
        cmd.execute(
            verbose=False,
            build_dir=tmp_path,
            organization_dir=data.PROJECT_NO_COGNITE_MODULES,
            selected=None,
            build_env_name="dev",
            no_clean=False,
        )

        assert not cmd.warning_list, f"No warnings should be raised. Got warnings: {cmd.warning_list}"
        # There are two transformations in the project, expect two transformation files
        transformation_files = [
            f
            for f in (tmp_path / "transformations").iterdir()
            if f.is_file() and TransformationCRUD.is_supported_file(f)
        ]
        assert len(transformation_files) == 2

    def test_build_complete_org_without_warnings(
        self,
        tmp_path: Path,
        env_vars_with_client: EnvironmentVariables,
    ) -> None:
        cmd = BuildCommand(silent=True, skip_tracking=True)
        with patch.dict(
            os.environ,
            {"CDF_PROJECT": env_vars_with_client.CDF_PROJECT, "CDF_CLUSTER": env_vars_with_client.CDF_CLUSTER},
        ):
            cmd.execute(
                verbose=False,
                build_dir=tmp_path / "build",
                organization_dir=data.COMPLETE_ORG,
                selected=None,
                build_env_name="dev",
                no_clean=False,
            )

        assert not cmd.warning_list, (
            f"No warnings should be raised. Got {len(cmd.warning_list)} warnings: {cmd.warning_list}"
        )

    def test_build_no_warnings_when_space_exists_in_cdf(
        self, env_vars_with_client: EnvironmentVariables, toolkit_client_approval: ApprovalToolkitClient, tmp_path: Path
    ) -> None:
        my_group = """name: gp_trigger_issue
sourceId: '1234567890123456789'
capabilities:
- dataModelInstancesAcl:
    actions:
    - READ
    scope:
      spaceIdScope:
        spaceIds:
        - existing-space
"""
        filepath = tmp_path / "my_org" / "modules" / "my_module" / "auth" / "my.Group.yaml"
        filepath.parent.mkdir(parents=True, exist_ok=True)
        filepath.write_text(my_group)

        # Simulate that the space exists in CDF
        toolkit_client_approval.append(Space, Space("existing-space", False, 1, 1, None, None))
        cmd = BuildCommand(silent=True, skip_tracking=True)
        with patch.dict(
            os.environ,
            {"CDF_PROJECT": env_vars_with_client.CDF_PROJECT, "CDF_CLUSTER": env_vars_with_client.CDF_CLUSTER},
        ):
            cmd.execute(
                verbose=False,
                organization_dir=tmp_path / "my_org",
                build_dir=tmp_path / "build",
                selected=None,
                build_env_name=None,
                no_clean=False,
                client=toolkit_client_approval.mock_client,
                on_error="raise",
            )
        assert len(cmd.warning_list) == 0


class TestCheckYamlSemantics:
    def test_build_valid_read_int_version(self) -> None:
        cmd = BuildCommand(silent=True)
        raw_yaml = """destination:
  dataModel:
    destinationType: CogniteFile
    externalId: MyModel
    space: my_space
    version: 1_0_0
  instanceSpace: my_space
  type: instances
externalId: some_external_id
    """
        source_filepath = MagicMock(spec=Path)
        source_filepath.read_text.return_value = raw_yaml
        source_filepath.suffix = ".yaml"
        source_filepath.read_bytes.return_value = raw_yaml.encode("utf-8")

        source_files = cmd._replace_variables(
            [source_filepath], BuildVariables([]), TransformationCRUD.folder_name, Path("my_module"), verbose=False
        )
        assert len(source_files) == 1
        source_file = source_files[0]
        assert isinstance(source_file.loaded, dict)
        actual = DataModelId.load(source_file.loaded["destination"]["dataModel"])
        assert actual == DataModelId("my_space", "MyModel", "1_0_0")

    def test_track_module_build(self, tmp_path: Path) -> None:
        cmd = BuildCommand(print_warning=True, skip_tracking=True)
        cmd.run(
            lambda: cmd.build_modules(
                modules=ModuleDirectories.load(data.EXTERNAL_PACKAGE),
                build_dir=tmp_path,
                variables=BuildVariables([]),
                verbose=False,
            )
        )
        assert cmd._additional_tracking_info.package_ids == {"rmdm"}
        assert cmd._additional_tracking_info.module_ids == {"agent", "data_model"}

    def test_track_module_build_with_package_info(self, tmp_path: Path) -> None:
        cmd = BuildCommand(print_warning=True, skip_tracking=True)
        cmd.build_config(
            build_dir=tmp_path,
            organization_dir=data.EXTERNAL_PACKAGE,
            config=BuildConfigYAML(
                filepath=Path("config.dev.yaml"),
                environment=Environment(
                    name="dev", project="my_project", validation_type="dev", selected=["external_module_1"]
                ),
            ),
            packages=Packages.load(data.EXTERNAL_PACKAGE),
            clean=False,
            verbose=False,
            client=None,
            progress_bar=False,
            on_error="continue",
        )

        with open(tmp_path / "_build_environment.yaml") as file:
            _build_file = yaml.safe_load(file)
        assert _build_file is not None
        assert _build_file["read_modules"][0]["package_id"] == "rmdm"
        assert _build_file["read_modules"][0]["module_id"] == "data_model"



import re
import sys
import uuid
from collections import defaultdict
from collections.abc import Collection, Iterator, Sequence
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any, Literal, SupportsIndex, overload

from cognite_toolkit._cdf_tk.cruds._resource_cruds.transformation import TransformationCRUD
from cognite_toolkit._cdf_tk.data_classes._module_directories import ModuleLocation
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.feature_flags import Flags

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclass(frozen=True)
class BuildVariable:
    """This is an internal representation of a  build variable in a config.[env].file

    Args:
        key: The name of the variable.
        value: The value of the variable.
        is_selected: Whether the variable is selected by the user through Environment.selected
        location: The location for the variable which is used to determine the module(s) it belongs to

    """

    key: str
    value: str | int | float | bool | tuple[str | int | float | bool]
    is_selected: bool
    location: Path
    iteration: int | None = None

    @property
    def value_variable(self) -> str | int | float | bool | list[str | int | float | bool]:
        """Returns the value of the variable as a variable."""
        if isinstance(self.value, tuple):
            # Convert the tuple back to a list to make it JSON serializable
            return list(self.value)
        else:
            return self.value

    def dump(self) -> dict[str, Any]:
        return {
            "key": self.key,
            "value": self.value_variable,
            "is_selected": self.is_selected,
            "location": self.location.as_posix(),
        }

    @classmethod
    def load(cls, data: dict[str, Any]) -> Self:
        if isinstance(data["value"], list):
            # Convert the list to a tuple to make it hashable
            value = tuple(data["value"])
        else:
            value = data["value"]
        return cls(data["key"], value, data["is_selected"], Path(data["location"]))


class BuildVariables(tuple, Sequence[BuildVariable]):
    """This is an internal representation of the build variables in a config.[env].file

    The motivation for this class is to provide helper functions for the user to interact with the build variables.
    """

    # Subclassing tuple to make the class immutable. BuildVariables is expected to be initialized and
    # then used as a read-only object.
    def __new__(cls, collection: Collection[BuildVariable], source_path: Path | None = None) -> Self:
        # Need to override __new__ to as we are subclassing a tuple:
        #   https://stackoverflow.com/questions/1565374/subclassing-tuple-with-multiple-init-arguments
        return super().__new__(cls, tuple(collection))

    def __init__(self, collection: Collection[BuildVariable], source_path: Path | None = None) -> None:
        super().__init__()
        self.source_path = source_path

    @cached_property
    def selected(self) -> "BuildVariables":
        return BuildVariables([variable for variable in self if variable.is_selected])

    @classmethod
    def load_raw(
        cls,
        raw_variable: dict[str, Any],
        available_modules: set[Path],
        selected_modules: set[Path] | None = None,
        source_path: Path | None = None,
    ) -> Self:
        """Loads the variables from the user input."""
        variables = []
        to_check: list[tuple[Path, int | None, dict[str, Any]]] = [(Path(""), None, raw_variable)]
        while to_check:
            path, iteration, subdict = to_check.pop()
            for key, value in subdict.items():
                subpath = path / key
                if subpath in available_modules and isinstance(value, dict):
                    to_check.append((subpath, None, value))
                elif subpath in available_modules and isinstance(value, list):
                    if Flags.MODULE_REPEAT.is_enabled():
                        for no, module_variables in enumerate(value, 1):
                            if not isinstance(module_variables, dict):
                                raise ToolkitValueError(f"Variables under a module must be a dictionary: {subpath}.")
                            to_check.append((subpath, no, module_variables))
                    else:
                        raise ToolkitValueError(
                            f"Variables under a module cannot be a list: {subpath}. Please use a dictionary/mapping."
                        )
                elif isinstance(value, dict):
                    # Remove this check to support variables with dictionary values.
                    continue
                else:
                    hashable_values = tuple(value) if isinstance(value, list) else value
                    is_selected = selected_modules is None or path in selected_modules
                    variables.append(BuildVariable(key, hashable_values, is_selected, path, iteration))

        return cls(variables, source_path=source_path)

    @classmethod
    def load(cls, data: list[dict[str, Any]]) -> Self:
        """Loads the variables from a dictionary."""
        return cls([BuildVariable.load(variable) for variable in data])

    def get_module_variables(self, module: ModuleLocation) -> "list[BuildVariables]":
        """Gets the variables for a specific module."""
        variables_by_key_by_iteration: dict[int | None, dict[str, list[BuildVariable]]] = defaultdict(
            lambda: defaultdict(list)
        )
        for variable in self:
            if variable.location == module.relative_path or variable.location in module.parent_relative_paths:
                variables_by_key_by_iteration[variable.iteration][variable.key].append(variable)

        base_variables: dict[str, list[BuildVariable]] = variables_by_key_by_iteration.pop(None, {})
        variable_sets: list[dict[str, list[BuildVariable]]]
        if variables_by_key_by_iteration:
            # Combine each with the base variables
            variable_sets = []
            for _, variables_by_key in sorted(variables_by_key_by_iteration.items(), key=lambda x: x[0] or 0):
                for key, build_variable in base_variables.items():
                    variables_by_key[key].extend(build_variable)
                variable_sets.append(variables_by_key)
        else:
            variable_sets = [base_variables]

        return [
            BuildVariables(
                [
                    # We select the variable with the longest path to ensure that the most specific variable is selected
                    max(variables, key=lambda v: len(v.location.parts))
                    for variables in variable_set.values()
                ],
                source_path=self.source_path,
            )
            for variable_set in variable_sets
        ]

    @overload
    def replace(self, content: str, file_path: Path | None = None, use_placeholder: Literal[False] = False) -> str: ...

    @overload
    def replace(
        self, content: str, file_path: Path | None = None, use_placeholder: Literal[True] = True
    ) -> tuple[str, dict[str, BuildVariable]]: ...

    def replace(
        self, content: str, file_path: Path | None = None, use_placeholder: bool = False
    ) -> str | tuple[str, dict[str, BuildVariable]]:
        # Extract file suffix from path, default to .yaml if not provided
        file_suffix = file_path.suffix if file_path and file_path.suffix else ".yaml"

        variable_by_placeholder: dict[str, BuildVariable] = {}
        for variable in self:
            if not use_placeholder:
                replace = variable.value_variable
            else:
                replace = f"VARIABLE_{uuid.uuid4().hex[:8]}"
                variable_by_placeholder[replace] = variable

            _core_pattern = rf"{{{{\s*{variable.key}\s*}}}}"
            if file_suffix == ".sql":
                # For SQL files, convert lists to SQL-style tuples
                if isinstance(replace, list):
                    replace = self._format_list_as_sql_tuple(replace)
                content = re.sub(_core_pattern, str(replace), content)
            elif file_suffix in {".yaml", ".yml", ".json"}:
                # Check if this is a transformation file (ends with Transformation.yaml/yml)
                is_transformation_file = file_path is not None and f".{TransformationCRUD.kind}." in file_path.name
                # Check if variable is within a query field (SQL context)
                is_in_query_field = self._is_in_query_field(content, variable.key)

                # For lists in query fields, use SQL-style tuples
                # For transformation files, ensure SQL conversion is applied to query property variables
                if is_transformation_file and is_in_query_field and isinstance(replace, list):
                    replace = self._format_list_as_sql_tuple(replace)
                    # Use simple pattern for SQL context (no YAML quoting needed)
                    content = re.sub(_core_pattern, str(replace), content)
                else:
                    # Preserve data types for YAML
                    pattern = _core_pattern
                    if isinstance(replace, str) and (replace.isdigit() or replace.endswith(":")):
                        replace = f'"{replace}"'
                        pattern = rf"'{_core_pattern}'|{_core_pattern}|" + rf'"{_core_pattern}"'
                    elif replace is None:
                        replace = "null"
                    content = re.sub(pattern, str(replace), content)
            else:
                # For other file types, use simple string replacement
                content = re.sub(_core_pattern, str(replace), content)
        if use_placeholder:
            return content, variable_by_placeholder
        else:
            return content

    @staticmethod
    def _is_transformation_file(file_path: Path) -> bool:
        """Check if the file path indicates a transformation YAML file.

        Transformation files are YAML files in the "transformations" folder.

        Args:
            file_path: The file path to check

        Returns:
            True if the file is a transformation YAML file
        """
        # Check if path contains "transformations" folder and ends with .yaml/.yml
        path_str = file_path.as_posix().lower()
        return "transformations" in path_str and file_path.suffix.lower() in {".yaml", ".yml"}

    @staticmethod
    def _format_list_as_sql_tuple(replace: list[Any]) -> str:
        """Format a list as a SQL-style tuple string.

        Args:
            replace: The list to format

        Returns:
            SQL tuple string, e.g., "('A', 'B', 'C')" or "()" for empty lists
        """
        if not replace:
            # Empty list becomes empty SQL tuple
            return "()"
        else:
            # Format list as SQL tuple: ('A', 'B', 'C')
            formatted_items = []
            for item in replace:
                if item is None:
                    formatted_items.append("NULL")
                elif isinstance(item, str):
                    formatted_items.append(f"'{item}'")
                else:
                    formatted_items.append(str(item))
            return f"({', '.join(formatted_items)})"

    @staticmethod
    def _is_in_query_field(content: str, variable_key: str) -> bool:
        """Check if a variable is within a query field in YAML.

        Assumes query is a top-level property. This detects various YAML formats:
        - query: >-
        - query: |
        - query: "..."
        - query: ...
        """
        lines = content.split("\n")
        variable_pattern = rf"{{{{\s*{re.escape(variable_key)}\s*}}}}"
        in_query_field = False

        for line in lines:
            # Check if this line starts a top-level query field
            query_match = re.match(r"^query\s*:\s*(.*)$", line)
            if query_match:
                in_query_field = True
                query_content_start = query_match.group(1).strip()

                # Check if variable is on the same line as query: declaration
                if re.search(variable_pattern, line):
                    return True

                # If query content starts on same line (not a block scalar), check it
                if query_content_start and not query_content_start.startswith(("|", ">", "|-", ">-", "|+", ">+")):
                    if re.search(variable_pattern, query_content_start):
                        return True
                continue

            # Check if we're still in the query field
            if in_query_field:
                # If we hit another top-level property, we've exited the query field
                if re.match(r"^\w+\s*:", line):
                    in_query_field = False
                    continue

                # We're still in the query field, check for variable
                if re.search(variable_pattern, line):
                    return True

        return False

    # Implemented to get correct type hints
    def __iter__(self) -> Iterator[BuildVariable]:
        return super().__iter__()

    @overload
    def __getitem__(self, index: SupportsIndex) -> BuildVariable: ...

    @overload
    def __getitem__(self, index: slice) -> Self: ...

    def __getitem__(self, index: SupportsIndex | slice, /) -> "BuildVariable | BuildVariables":
        if isinstance(index, slice):
            return BuildVariables(super().__getitem__(index))
        return super().__getitem__(index)

    def dump(self) -> list[dict[str, Any]]:
        return [variable.dump() for variable in self]



"""
Approval test takes a snapshot of the results and then compare them to last run, ref https://approvaltests.com/,
and fails if they have changed.

If the changes are desired, you can update the snapshot by running `pytest tests/ --force-regen`.
"""

from collections import defaultdict
from collections.abc import Iterator
from pathlib import Path

import pytest
from pytest import MonkeyPatch

from cognite_toolkit._cdf_tk.commands import BuildCommand, CleanCommand, DeployCommand
from cognite_toolkit._cdf_tk.cruds import CRUDS_BY_FOLDER_NAME, Loader
from cognite_toolkit._cdf_tk.data_classes import ModuleDirectories
from cognite_toolkit._cdf_tk.feature_flags import Flags
from cognite_toolkit._cdf_tk.utils import humanize_collection, iterate_modules
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests.data import BUILDABLE_PACKAGE, COMPLETE_ORG, COMPLETE_ORG_ALPHA_FLAGS
from tests.test_unit.approval_client import ApprovalToolkitClient
from tests.test_unit.utils import mock_read_yaml_file

THIS_DIR = Path(__file__).resolve().parent
SNAPSHOTS_DIR = THIS_DIR / "test_build_deploy_snapshots"
SNAPSHOTS_DIR.mkdir(exist_ok=True)
SNAPSHOTS_DIR_CLEAN = THIS_DIR / "test_build_clean_snapshots"
SNAPSHOTS_DIR_CLEAN.mkdir(exist_ok=True)


def find_all_modules() -> Iterator[Path]:
    for module, _ in iterate_modules(BUILDABLE_PACKAGE):
        if module.name == "references":  # this particular module should never be built or deployed
            continue
        elif module.name == "search":
            # Not ready yet
            continue
        yield pytest.param(module, id=f"{module.parent.name}/{module.name}")


def mock_environments_yaml_file(module_path: Path, monkeypatch: MonkeyPatch) -> None:
    return mock_read_yaml_file(
        {
            "config.dev.yaml": {
                "environment": {
                    "name": "dev",
                    "project": "pytest-project",
                    "type": "dev",
                    "selected": [module_path.name],
                }
            }
        },
        monkeypatch,
        modify=True,
    )


@pytest.mark.parametrize("module_path", list(find_all_modules()))
def test_build_deploy_module(
    module_path: Path,
    build_tmp_path: Path,
    monkeypatch: MonkeyPatch,
    toolkit_client_approval: ApprovalToolkitClient,
    env_vars_with_client: EnvironmentVariables,
    buildable_modules: Path,
    data_regression,
) -> None:
    BuildCommand(skip_tracking=True, silent=True).execute(
        verbose=False,
        organization_dir=buildable_modules,
        build_dir=build_tmp_path,
        selected=[module_path.name],
        build_env_name="dev",
        no_clean=False,
        client=env_vars_with_client.get_client(),
        on_error="raise",
    )

    DeployCommand(skip_tracking=True, silent=True).deploy_build_directory(
        env_vars=env_vars_with_client,
        build_dir=build_tmp_path,
        build_env_name="dev",
        drop=True,
        dry_run=False,
        include=[],
        verbose=False,
        drop_data=True,
        force_update=False,
    )

    not_mocked = toolkit_client_approval.not_mocked_calls()
    assert not not_mocked, (
        f"The following APIs have been called without being mocked: {not_mocked}, "
        "Please update the list _API_RESOURCES in tests/approval_client.py"
    )

    dump = toolkit_client_approval.dump()
    data_regression.check(dump, fullpath=SNAPSHOTS_DIR / f"{module_path.name}.yaml")

    for group_calls in toolkit_client_approval.auth_create_group_calls():
        lost_capabilities = group_calls.capabilities_all_calls - group_calls.last_created_capabilities
        assert not lost_capabilities, (
            f"The group {group_calls.name!r} has lost the capabilities: {', '.join(lost_capabilities)}"
        )


@pytest.mark.parametrize("module_path", list(find_all_modules()))
def test_build_deploy_with_dry_run(
    module_path: Path,
    build_tmp_path: Path,
    monkeypatch: MonkeyPatch,
    toolkit_client_approval: ApprovalToolkitClient,
    env_vars_with_client: EnvironmentVariables,
    buildable_modules: Path,
) -> None:
    mock_environments_yaml_file(module_path, monkeypatch)

    BuildCommand(skip_tracking=True, silent=True).execute(
        organization_dir=buildable_modules,
        build_dir=build_tmp_path,
        selected=None,
        build_env_name="dev",
        no_clean=False,
        client=env_vars_with_client.get_client(),
        on_error="raise",
        verbose=False,
    )
    DeployCommand(skip_tracking=True, silent=True).deploy_build_directory(
        env_vars=env_vars_with_client,
        build_dir=build_tmp_path,
        build_env_name="dev",
        drop=True,
        dry_run=True,
        include=[],
        verbose=False,
        drop_data=True,
        force_update=False,
    )

    create_result = toolkit_client_approval.create_calls()
    assert not create_result, f"No resources should be created in dry run: got these calls: {create_result}"
    delete_result = toolkit_client_approval.delete_calls()
    assert not delete_result, f"No resources should be deleted in dry run: got these calls: {delete_result}"


@pytest.mark.parametrize("module_path", list(find_all_modules()))
def test_init_build_clean(
    module_path: Path,
    build_tmp_path: Path,
    monkeypatch: MonkeyPatch,
    toolkit_client_approval: ApprovalToolkitClient,
    env_vars_with_client: EnvironmentVariables,
    buildable_modules: Path,
    data_regression,
) -> None:
    mock_environments_yaml_file(module_path, monkeypatch)

    BuildCommand(silent=True, skip_tracking=True).execute(
        verbose=False,
        organization_dir=buildable_modules,
        build_dir=build_tmp_path,
        selected=None,
        no_clean=False,
        client=env_vars_with_client.get_client(),
        build_env_name="dev",
        on_error="raise",
    )
    CleanCommand(silent=True, skip_tracking=True).execute(
        env_vars=env_vars_with_client,
        build_dir=build_tmp_path,
        build_env_name="dev",
        dry_run=False,
        include=None,
        module_str=module_path.name,
        all_modules=True,
        verbose=False,
    )

    not_mocked = toolkit_client_approval.not_mocked_calls()
    assert not not_mocked, (
        f"The following APIs have been called without being mocked: {not_mocked}, "
        "Please update the list _API_RESOURCES in tests/approval_client.py"
    )
    dump = toolkit_client_approval.dump()
    data_regression.check(dump, fullpath=SNAPSHOTS_DIR_CLEAN / f"{module_path.name}.yaml")


TEST_CASES = [COMPLETE_ORG]
if Flags.GRAPHQL.is_enabled():
    TEST_CASES.append(COMPLETE_ORG_ALPHA_FLAGS)


@pytest.mark.parametrize("organization_dir", TEST_CASES, ids=[path.name for path in TEST_CASES])
def test_build_deploy_complete_org(
    organization_dir: Path,
    build_tmp_path: Path,
    monkeypatch: MonkeyPatch,
    toolkit_client_approval: ApprovalToolkitClient,
    env_vars_with_client: EnvironmentVariables,
    data_regression,
) -> None:
    BuildCommand(silent=True, skip_tracking=True).execute(
        verbose=False,
        organization_dir=organization_dir,
        build_dir=build_tmp_path,
        selected=None,
        build_env_name="dev",
        no_clean=False,
        client=env_vars_with_client.get_client(),
        on_error="raise",
    )
    DeployCommand(silent=True, skip_tracking=True).deploy_build_directory(
        env_vars=env_vars_with_client,
        build_dir=build_tmp_path,
        build_env_name="dev",
        drop=True,
        drop_data=True,
        dry_run=False,
        force_update=False,
        include=None,
        verbose=False,
    )

    not_mocked = toolkit_client_approval.not_mocked_calls()
    assert not not_mocked, (
        f"The following APIs have been called without being mocked: {not_mocked}, "
        "Please update the list _API_RESOURCES in tests/approval_client.py"
    )

    dump = toolkit_client_approval.dump()
    data_regression.check(dump, fullpath=SNAPSHOTS_DIR / f"{organization_dir.name}.yaml")

    for group_calls in toolkit_client_approval.auth_create_group_calls():
        lost_capabilities = group_calls.capabilities_all_calls - group_calls.last_created_capabilities
        assert not lost_capabilities, (
            f"The group {group_calls.name!r} has lost the capabilities: {', '.join(lost_capabilities)}"
        )


def test_complete_org_is_complete() -> None:
    modules = ModuleDirectories.load(COMPLETE_ORG)
    used_loader_by_folder_name: dict[str, set[type[Loader]]] = defaultdict(set)

    for module in modules:
        for resource_folder, files in module.source_paths_by_resource_folder.items():
            for loader in CRUDS_BY_FOLDER_NAME[resource_folder]:
                if any(loader.is_supported_file(file) for file in files):
                    used_loader_by_folder_name[resource_folder].add(loader)
    alpha_modules = ModuleDirectories.load(COMPLETE_ORG_ALPHA_FLAGS)
    for module in alpha_modules:
        for resource_folder, files in module.source_paths_by_resource_folder.items():
            for loader in CRUDS_BY_FOLDER_NAME[resource_folder]:
                if any(loader.is_supported_file(file) for file in files):
                    used_loader_by_folder_name[resource_folder].add(loader)

    unused_loaders = {
        loader
        for folder, loaders in CRUDS_BY_FOLDER_NAME.items()
        for loader in loaders
        if loader not in used_loader_by_folder_name[folder]
    }

    # If this assertion fails, it means that the complete_org is not complete.
    # This typically happens when you have just added a new loader and forgotten to add
    # example data for the new resource type in the tests/data/complete_org directory.
    assert not unused_loaders, (
        f"The following {len(unused_loaders)} loaders are not used: {humanize_collection([loader.__name__ for loader in unused_loaders])}"
    )



import pytest
import responses

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.charts import Chart, ChartList, ChartWrite
from cognite_toolkit._cdf_tk.client.data_classes.charts_data import ChartData, ChartSettings, UserInfo
from tests.test_unit.utils import FakeCogniteResourceGenerator

CHART = Chart(
    external_id="chart",
    created_time=1,
    last_updated_time=2,
    visibility="PUBLIC",
    data=ChartData(
        version=1,
        name="TestNew",
        date_from="2025-04-26T22:00:00.000Z",
        date_to="2025-05-27T21:59:59.999Z",
        user_info=UserInfo(id="toolkit_test_user", email="support@cognite.com", display_name="Toolkit Test User"),
        settings=ChartSettings(
            show_y_axis=True,
            show_min_max=True,
            show_gridlines=True,
            merge_units=True,
        ),
    ),
    owner_id="toolkit_test_user",
)


class TestChartAPI:
    @pytest.mark.parametrize(
        "items, expected_return_cls",
        [
            pytest.param(
                CHART.as_write(),
                Chart,
                id="Single ChartWrite",
            ),
            pytest.param(
                [CHART.as_write()],
                ChartList,
                id="List of ChartWrite",
            ),
        ],
    )
    def test_upsert_return_type(
        self,
        items: ChartWrite | list[ChartWrite],
        expected_return_cls: type,
        toolkit_config: ToolkitClientConfig,
    ) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        url = f"{toolkit_config.base_url}/apps/v1/projects/{toolkit_config.project}/storage/charts/charts"

        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.PUT,
                url,
                status=200,
                json={"items": [CHART.dump()]},
            )
            result = client.charts.upsert(items)

        assert isinstance(result, expected_return_cls)

    @pytest.mark.parametrize(
        "external_id, expected_return_cls",
        [
            pytest.param(
                CHART.external_id,
                Chart,
                id="Single Chart",
            ),
            pytest.param(
                [CHART.external_id],
                ChartList,
                id="List of Chart",
            ),
        ],
    )
    def test_retrieve_return_type(
        self, external_id: str | list[str], expected_return_cls: type, toolkit_config: ToolkitClientConfig
    ) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        url = f"{toolkit_config.base_url}/apps/v1/projects/{toolkit_config.project}/storage/charts/charts/byids"
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={"items": [CHART.dump()]},
            )
            result = client.charts.retrieve(external_id=external_id)

        assert isinstance(result, expected_return_cls)


class TestChartDTOs:
    def test_chart_data_changed(self) -> None:
        """The ChartData is frontend of the Chart object, and it is not enforced in any way by the backend API.
        Thus, it can change completely without any notice. This tests ensures that whatever the ChartData is,
        the serialization and deserialization works correctly.
        """
        chart_data = {
            "this": "is",
            "completely": {
                "changed": ["compared", "to", "the", "previous", "version"],
            },
            "and": 123,
            "itAlso": {"hasSomeNumbers": [1, 2, 3, 4, 5]},
        }
        loaded = ChartData._load(chart_data)
        dumped = loaded.dump(camel_case=False)

        assert dumped == chart_data, f"Expected {chart_data}, but got {dumped}"

    def test_serialize_deserialize_chart(self) -> None:
        """Test that Chart can be serialized and deserialized correctly."""
        instance = FakeCogniteResourceGenerator(seed=42).create_instance(ChartData)
        instance._unknown_fields = {"extraField": "extraValue"}

        dumped = instance.dump(camel_case=True)

        loaded = ChartData._load(dumped)

        assert loaded.dump() == instance.dump(), f"Expected {instance}, but got {loaded}"



import sys
from abc import ABC, abstractmethod
from collections import UserList
from collections.abc import Collection
from pathlib import Path
from typing import Literal, TypeVar

from cognite.client.utils._text import to_camel_case
from pydantic import BaseModel

from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.fileio import CSVReader, SchemaColumn
from cognite_toolkit._cdf_tk.validation import T_BaseModel, instantiate_class

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class ModelList(UserList[T_BaseModel], ABC):
    def __init__(
        self,
        collection: Collection[T_BaseModel] | None = None,
        invalid_rows: dict[int, ResourceFormatWarning] | None = None,
        unexpected_columns: set[str] | None = None,
    ) -> None:
        super().__init__(collection or [])
        self.invalid_rows = invalid_rows or {}
        self.unexpected_columns = unexpected_columns or set()

    @classmethod
    @abstractmethod
    def _get_base_model_cls(cls) -> type[T_BaseModel]:
        raise NotImplementedError()

    @classmethod
    def _required_header_names(cls) -> set[str]:
        model_cls = cls._get_base_model_cls()
        return {field_.alias or field_id for field_id, field_ in model_cls.model_fields.items() if field_.is_required()}

    @classmethod
    def _optional_header_names(cls) -> set[str]:
        model_cls = cls._get_base_model_cls()
        return {
            field_.alias or field_id for field_id, field_ in model_cls.model_fields.items() if not field_.is_required()
        }

    @classmethod
    def read_csv_file(cls, filepath: Path) -> "Self":
        # We only need to read one row to get the header
        schema = CSVReader.sniff_schema(filepath, sniff_rows=1)
        cls._validate_schema(schema)
        unexpected_columns = {col.name for col in schema} - cls._required_header_names() - cls._optional_header_names()
        reader = CSVReader(input_file=filepath)
        items: list[T_BaseModel] = []
        invalid_rows: dict[int, ResourceFormatWarning] = {}
        model_cls = cls._get_base_model_cls()
        for row_no, row in enumerate(reader.read_chunks_unprocessed(), start=1):
            result = instantiate_class(row, model_cls, filepath)
            if isinstance(result, model_cls):
                items.append(result)
            elif isinstance(result, ResourceFormatWarning):
                invalid_rows[row_no] = result
            else:
                raise TypeError(f"Unexpected result type: {type(result)}")

        return cls(items, invalid_rows, unexpected_columns)

    @classmethod
    def _validate_schema(cls, schema: list[SchemaColumn]) -> None:
        actual = {col.name for col in schema}
        if missing_columns := cls._required_header_names() - actual:
            raise ToolkitValueError(f"Missing required columns: {humanize_collection(missing_columns)}")


T_ModelList = TypeVar("T_ModelList", bound=ModelList)


class InstanceIdRow(BaseModel, alias_generator=to_camel_case):
    space: str
    external_id: str
    instance_type: Literal["node", "edge"] = "node"


class InstanceIdCSVList(ModelList[InstanceIdRow]):
    @classmethod
    def _get_base_model_cls(cls) -> type[InstanceIdRow]:
        return InstanceIdRow



from typing import Annotated

from pydantic import Field, TypeAdapter

from ._asset_centric import AssetCentricFileSelector, AssetCentricSelector, AssetSubtreeSelector, DataSetSelector
from ._base import DataSelector
from ._canvas import CanvasSelector
from ._charts import AllChartsSelector, ChartOwnerSelector, ChartSelector
from ._datapoints import (
    DataPointsFileSelector,
    ExternalIdColumn,
    InstanceColumn,
    InternalIdColumn,
    TimeSeriesColumn,
)
from ._instances import (
    InstanceFileSelector,
    InstanceSelector,
    InstanceSpaceSelector,
    InstanceViewSelector,
    SelectedView,
)
from ._raw import RawTableSelector, SelectedTable

Selector = Annotated[
    RawTableSelector
    | InstanceViewSelector
    | InstanceFileSelector
    | InstanceSpaceSelector
    | AllChartsSelector
    | ChartOwnerSelector
    | AssetSubtreeSelector
    | AssetCentricFileSelector
    | DataSetSelector
    | DataPointsFileSelector,
    Field(discriminator="type"),
]

SelectorAdapter: TypeAdapter[Selector] = TypeAdapter(Selector)


__all__ = [
    "AllChartsSelector",
    "AssetCentricFileSelector",
    "AssetCentricSelector",
    "AssetSubtreeSelector",
    "CanvasSelector",
    "ChartOwnerSelector",
    "ChartSelector",
    "DataPointsFileSelector",
    "DataSelector",
    "DataSetSelector",
    "ExternalIdColumn",
    "InstanceColumn",
    "InstanceFileSelector",
    "InstanceSelector",
    "InstanceSpaceSelector",
    "InstanceViewSelector",
    "InternalIdColumn",
    "RawTableSelector",
    "SelectedTable",
    "SelectedView",
    "Selector",
    "SelectorAdapter",
    "TimeSeriesColumn",
]



from pathlib import Path
from unittest.mock import MagicMock

import pytest
import responses
import yaml
from cognite.client.credentials import OAuthClientCredentials
from cognite.client.data_classes import (
    ClientCredentials,
    Function,
    FunctionSchedule,
    FunctionScheduleWrite,
    FunctionScheduleWriteList,
    FunctionWrite,
)
from cognite.client.data_classes.capabilities import FilesAcl, FunctionsAcl
from cognite.client.exceptions import CogniteAPIError

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.cruds import FunctionCRUD, FunctionScheduleCRUD, ResourceWorker
from cognite_toolkit._cdf_tk.exceptions import ResourceCreationError, ToolkitRequiredValueError
from cognite_toolkit._cdf_tk.utils import calculate_directory_hash, calculate_secure_hash
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests.data import LOAD_DATA
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestFunctionLoader:
    def test_load_functions(self, env_vars_with_client: EnvironmentVariables) -> None:
        loader = FunctionCRUD.create_loader(env_vars_with_client.get_client(), LOAD_DATA)

        raw_list = loader.load_resource_file(
            LOAD_DATA / "functions" / "1.my_functions.yaml", env_vars_with_client.dump()
        )

        assert len(raw_list) == 2

    def test_load_function(self, env_vars_with_client: EnvironmentVariables) -> None:
        loader = FunctionCRUD.create_loader(env_vars_with_client.get_client(), LOAD_DATA)

        raw_list = loader.load_resource_file(
            LOAD_DATA / "functions" / "1.my_function.yaml", env_vars_with_client.dump()
        )
        loaded = loader.load_resource(raw_list[0], is_dry_run=False)

        assert isinstance(loaded, FunctionWrite)

    def test_update_secrets(
        self, env_vars_with_client: EnvironmentVariables, toolkit_client_approval: ApprovalToolkitClient, tmp_path: Path
    ) -> None:
        local_yaml = """name: my_function
externalId: my_function
secrets:
    secret1: value1
    secret2: value2
        """
        cdf_function = Function(
            name="my_function",
            external_id="my_function",
            file_id=123,
            status="Ready",
            metadata={
                FunctionCRUD._MetadataKey.function_hash: FunctionCRUD._create_hash_values(tmp_path / "my_function"),
                FunctionCRUD._MetadataKey.secret_hash: calculate_secure_hash(
                    {
                        "secret1": "value1",
                        "secret2": "value2",
                    }
                ),
            },
            secrets={
                # The API returns secrets masked
                "secret1": "***",
                "secret2": "***",
            },
        )
        toolkit_client_approval.append(Function, cdf_function)

        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = local_yaml
        filepath.parent.name = FunctionCRUD.folder_name

        worker = ResourceWorker(FunctionCRUD.create_loader(env_vars_with_client.get_client(), tmp_path), "deploy")
        resources = worker.prepare_resources([filepath])

        assert {
            "create": len(resources.to_create),
            "update": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 0, "update": 0, "delete": 0, "unchanged": 1}

        toolkit_client_approval.clear_cdf_resources(Function)
        cdf_function.metadata[FunctionCRUD._MetadataKey.secret_hash] = calculate_secure_hash(
            {
                "secret1": "value1",
                "secret2": "updated_value2",
            }
        )
        toolkit_client_approval.append(Function, cdf_function)
        resources = worker.prepare_resources([filepath])

        assert {
            "create": len(resources.to_create),
            "update": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 1, "update": 0, "delete": 1, "unchanged": 0}

    def test_dump_index_url_set(self, env_vars_with_client: EnvironmentVariables, tmp_path: Path) -> None:
        local_dict = FunctionWrite(
            name="my_function",
            file_id=123,
            external_id="my_function",
            index_url="http://my-index-url",
        ).dump()
        cdf_function = Function(
            name="my_function",
            file_id=123,
            external_id="my_function",
            metadata={
                FunctionCRUD._MetadataKey.function_hash: calculate_directory_hash(
                    tmp_path / "my_function", exclude_prefixes={".DS_Store"}
                ),
            },
        )
        loader = FunctionCRUD.create_loader(env_vars_with_client.get_client(), tmp_path)

        dumped = loader.dump_resource(cdf_function, local_dict)

        assert "indexUrl" in dumped
        assert dumped["indexUrl"] == "http://my-index-url"

    def test_get_function_required_capabilities(self, env_vars_with_client: EnvironmentVariables) -> None:
        loader = FunctionCRUD.create_loader(env_vars_with_client.get_client(), None)
        loader.data_set_id_by_external_id = {"function1": 123, "function2": 456}

        # Mock data
        items = [
            FunctionWrite(external_id="function1", name="Function 1", file_id=1001),
            FunctionWrite(external_id="function2", name="Function 2", file_id=1002),
            FunctionWrite(external_id="function3", name="Function 3", file_id=1003),
        ]

        # Test read-only mode
        read_capabilities = loader.get_function_required_capabilities(items, read_only=True)
        assert len(read_capabilities) == 2
        assert isinstance(read_capabilities[0], FunctionsAcl)
        assert isinstance(read_capabilities[1], FilesAcl)
        assert read_capabilities[0].actions == [FunctionsAcl.Action.Read]
        assert read_capabilities[1].actions == [FilesAcl.Action.Read]
        assert isinstance(read_capabilities[1].scope, FilesAcl.Scope.DataSet)
        assert sorted(read_capabilities[1].scope.ids) == [123, 456]

        # Test write mode
        write_capabilities = loader.get_function_required_capabilities(items, read_only=False)
        assert len(write_capabilities) == 2
        assert write_capabilities[0].actions == [FunctionsAcl.Action.Read, FunctionsAcl.Action.Write]
        assert write_capabilities[1].actions == [FilesAcl.Action.Read, FilesAcl.Action.Write]
        assert isinstance(write_capabilities[1].scope, FilesAcl.Scope.DataSet)
        assert sorted(write_capabilities[1].scope.ids) == [123, 456]

    def test_get_function_required_capabilities_empty(self, env_vars_with_client: EnvironmentVariables) -> None:
        loader = FunctionCRUD.create_loader(env_vars_with_client.get_client(), None)
        capabilities = loader.get_function_required_capabilities([], read_only=False)
        assert capabilities == []

    def test_get_function_required_capabilities_no_datasets(self, env_vars_with_client: EnvironmentVariables) -> None:
        loader = FunctionCRUD.create_loader(env_vars_with_client.get_client(), None)
        items = [
            FunctionWrite(external_id="function1", name="Function 1", file_id=1001),
            FunctionWrite(external_id="function2", name="Function 2", file_id=1002),
        ]

        capabilities = loader.get_function_required_capabilities(items, read_only=False)
        assert len(capabilities) == 2
        assert isinstance(capabilities[1].scope, FilesAcl.Scope.All)


class TestFunctionScheduleLoader:
    def test_credentials_missing_raise(self) -> None:
        schedule = dict(
            name="daily-8am-utc",
            functionExternalId="fn_example_repeater",
            cronExpression="0 8 * * *",
        )
        config = MagicMock(spec=ToolkitClientConfig)
        config.is_strict_validation = True
        config.credentials = OAuthClientCredentials(
            client_id="toolkit-client-id",
            client_secret="toolkit-client-secret",
            token_url="https://cognite.com/token",
            scopes=["USER_IMPERSONATION"],
        )
        with monkeypatch_toolkit_client() as client:
            client.config = config
            loader = FunctionScheduleCRUD.create_loader(client)

        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = yaml.dump(schedule)
        with pytest.raises(ToolkitRequiredValueError):
            loader.load_resource_file(filepath, {})
        client.config.is_strict_validation = False
        filepath.read_text.return_value = yaml.dump(schedule)
        local = loader.load_resource_file(filepath, {})[0]
        credentials = loader.authentication_by_id[loader.get_id(local)]
        assert credentials.client_id == "toolkit-client-id"
        assert credentials.client_secret == "toolkit-client-secret"

    def test_credentials_unchanged_changed(self) -> None:
        local_content = """name: daily-8am-utc
functionExternalId: fn_example_repeater
cronExpression: 0 8 * * *
description: Run the function every day at 8am UTC
authentication:
  clientId: my-client-id
  clientSecret: my-client-secret
"""
        auth_dict = yaml.CSafeLoader(local_content).get_data()["authentication"]
        auth_hash = calculate_secure_hash(auth_dict, shorten=True)

        with monkeypatch_toolkit_client() as client:
            cdf_schedule = FunctionSchedule(
                id=123,
                name="daily-8am-utc",
                function_external_id="fn_example_repeater",
                cron_expression="0 8 * * *",
                description=f"Run the function every day at 8am UTC {FunctionScheduleCRUD._hash_key}: {auth_hash}",
                cognite_client=client,
            )
            # The as_write method looks up the input data.
            client.functions.schedules.get_input_data.return_value = None
            loader = FunctionScheduleCRUD(client, None, None)

        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = local_content
        local_dumped = loader.load_resource_file(filepath, {})[0]
        cdf_dumped = loader.dump_resource(cdf_schedule, local_dumped)

        assert cdf_dumped == local_dumped

        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = local_content.replace("my-client-secret", "my-client-secret-changed")
        local_dumped = loader.load_resource_file(filepath, {})[0]
        cdf_dumped = loader.dump_resource(cdf_schedule, local_dumped)

        assert cdf_dumped != local_dumped

    @pytest.mark.parametrize(
        "client_id,client_secret, expected_error",
        [
            pytest.param(
                "invalid-client-id",
                "invalid-client-secret",
                CogniteAPIError(message="Invalid client credentials", code=400),
                id="Invalid credentials no extra hint",
            ),
            pytest.param(
                "123789109801",
                "${MY_SUPER_SECRET}",
                ResourceCreationError(
                    "Failed to create Function Schedule FunctionScheduleID(function_external_id='fn_example_repeater', name='daily-8am-utc'): The environment variable is not set: MY_SUPER_SECRET"
                ),
                id="Invalid credentials missing envrionment variable",
            ),
        ],
    )
    def test_create_error_message_invalid_credentials(
        self, client_id: str, client_secret: str, expected_error: Exception, toolkit_config: ToolkitClientConfig
    ) -> None:
        config = toolkit_config
        with responses.RequestsMock() as rsps:
            rsps.add(
                rsps.POST,
                config.create_api_url("functions/byids"),
                status=200,
                json={
                    "items": [
                        {
                            "id": 1,
                            "name": "fn_example_repeater",
                            "externalId": "fn_example_repeater",
                            "status": "Ready",
                            "fileId": 101,
                        }
                    ]
                },
            )
            rsps.add(
                rsps.POST,
                config.create_api_url("/sessions"),
                status=400,
                json={"error": {"message": "Invalid client credentials"}},
            )
            client = ToolkitClient(toolkit_config)
            loader = FunctionScheduleCRUD(client, None, None)
            schedule = FunctionScheduleWrite(
                name="daily-8am-utc",
                function_external_id="fn_example_repeater",
                cron_expression="0 8 * * *",
            )
            id_ = loader.get_id(schedule)
            loader.authentication_by_id[id_] = ClientCredentials(
                client_id=client_id,
                client_secret=client_secret,
            )
            with pytest.raises(type(expected_error)) as exc_val:
                loader.create(FunctionScheduleWriteList([schedule]))

            assert str(expected_error) in str(exc_val.value)

    def test_create_function_schedule_missing_function(self, toolkit_config: ToolkitClientConfig) -> None:
        config = toolkit_config
        with responses.RequestsMock() as rsps:
            rsps.add(
                rsps.POST,
                config.create_api_url("functions/byids"),
                status=200,
                json={"items": []},
            )
            client = ToolkitClient(toolkit_config)
            loader = FunctionScheduleCRUD(client, None, None)
            schedule = FunctionScheduleWrite(
                name="daily-8am-utc",
                function_external_id="fn_non_existent_function",
                cron_expression="0 8 * * *",
            )
            with pytest.raises(ResourceCreationError) as exc_val:
                loader.create(FunctionScheduleWriteList([schedule]))

            assert (
                "Failed to create function schedule "
                "FunctionScheduleID(function_external_id='fn_non_existent_function', name='daily-8am-utc'). "
                "Could not find function 'fn_non_existent_function'"
            ) in str(exc_val.value)



from collections.abc import Sequence
from typing import overload
from urllib.parse import urljoin

from cognite.client._api_client import APIClient
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.charts import (
    Chart,
    ChartList,
    ChartWrite,
    Visibility,
)
from cognite_toolkit._cdf_tk.utils.collection import chunker


class ChartsAPI(APIClient):
    _RESOURCE_PATH = "/storage/charts/charts"

    def _get_base_url_with_base_path(self) -> str:
        """
        This method is overridden to provide the correct base path for the Search Configurations API.
        This method in base class APIClient appends /api/{api_version}/ to the base URL,
        but for Charts API, we need a different path structure.
        """
        base_path = ""
        if self._api_version:
            base_path = f"/apps/{self._api_version}/projects/{self._config.project}"
        return urljoin(self._config.base_url, base_path)

    @overload
    def upsert(self, items: ChartWrite) -> Chart: ...

    @overload
    def upsert(self, items: Sequence[ChartWrite]) -> ChartList: ...

    def upsert(self, items: ChartWrite | Sequence[ChartWrite]) -> Chart | ChartList:
        """Upsert one or more charts in CDF.

        Args:
            items (ChartWrite | Sequence[ChartWrite]): Chart(s) to upsert. Can be a single ChartWrite or a sequence of ChartWrite.

        Returns:
            Chart | ChartList: List of upserted charts if multiple items are provided, otherwise a single Chart.
        """
        item_sequence = items if isinstance(items, Sequence) else [items]
        result = ChartList([])
        # We are avoiding concurrency here as the Charts backend it not necessarily designed for it.
        for chunk in chunker(item_sequence, self._CREATE_LIMIT):
            body = {"items": [chunk_item.as_write().dump() for chunk_item in chunk]}
            response = self._put(
                url_path=self._RESOURCE_PATH,
                json=body,
            )
            result.extend([Chart._load(item, cognite_client=self._cognite_client) for item in response.json()["items"]])

        if not result:
            raise ValueError("No charts were upserted. This may indicate an issue with the upsert endpoint.")

        if isinstance(items, ChartWrite):
            if len(result) != 1:
                raise ValueError(
                    "Expected a single chart to be returned, but multiple charts were returned. "
                    "This may indicate an issue with the upsert operation."
                )
            return result[0]
        elif len(result) != len(items):
            raise ValueError(
                "The number of upserted charts does not match the number of input charts. "
                "This may indicate an issue with the upsert endpoint."
            )
        return result

    @overload
    def retrieve(self, external_id: str) -> Chart: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str]) -> ChartList: ...

    def retrieve(self, external_id: str | SequenceNotStr[str]) -> Chart | ChartList:
        """Retrieve one or more charts by their external ID.

        Args:
            external_id (str | Sequence[str]): External ID(s) of the chart(s) to retrieve.

        Returns:
            Chart | ChartList: A single Chart if a single external ID is provided, otherwise a ChartList.
        """
        return self._retrieve_multiple(
            list_cls=ChartList, resource_cls=Chart, identifiers=IdentifierSequence.load(external_ids=external_id)
        )

    def delete(self, external_id: str | SequenceNotStr[str]) -> None:
        """Delete one or more charts by their external ID.

        Args:
            external_id (str | Sequence[str]): External ID(s) of the chart(s) to delete.
        """
        self._delete_multiple(
            identifiers=IdentifierSequence.load(external_ids=external_id),
            wrap_ids=True,
        )

    def list(self, visibility: Visibility | None = None, is_owned: bool | None = None) -> ChartList:
        """List charts based on visibility and ownership.

        Args:
            visibility (Visibility): Visibility of the charts to list, either 'PUBLIC' or 'PRIVATE'.
            is_owned (bool): Whether to list only owned charts.

        Returns:
            ChartList: List of charts matching the criteria.
        """
        filter_: dict[str, str | bool] = {}
        if visibility is not None:
            filter_["visibility"] = visibility.upper()
        if is_owned is not None:
            filter_["isOwned"] = is_owned

        response = self._post(
            url_path=self._RESOURCE_PATH + "/list",
            json={"filter": filter_} if filter_ else {},
        )
        return ChartList._load(response.json()["items"], cognite_client=self._cognite_client)



from pathlib import Path
from unittest.mock import MagicMock

import pytest
from cognite.client.credentials import OAuthClientCredentials
from cognite.client.data_classes import WorkflowTrigger
from cognite.client.data_classes.workflows import (
    SubworkflowReferenceParameters,
    WorkflowDefinitionUpsert,
    WorkflowScheduledTriggerRule,
    WorkflowTask,
    WorkflowVersionId,
    WorkflowVersionUpsert,
)

from cognite_toolkit._cdf_tk.client import ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.cruds import WorkflowTriggerCRUD, WorkflowVersionCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitCycleError, ToolkitRequiredValueError
from cognite_toolkit._cdf_tk.utils import calculate_secure_hash


class TestWorkflowTriggerLoader:
    def test_credentials_missing_raise(self) -> None:
        trigger_content = """externalId: daily-8am-utc
triggerRule:
    triggerType: schedule
    cronExpression: 0 8 * * *
workflowExternalId: wf_example_repeater
workflowVersion: v1
        """
        trigger_file = MagicMock(spec=Path)
        trigger_file.read_text.return_value = trigger_content

        config = MagicMock(spec=ToolkitClientConfig)
        config.is_strict_validation = True
        config.credentials = OAuthClientCredentials(
            client_id="toolkit-client-id",
            client_secret="toolkit-client-secret",
            token_url="https://cognite.com/token",
            scopes=["USER_IMPERSONATION"],
        )
        with monkeypatch_toolkit_client() as client:
            client.config = config
            loader = WorkflowTriggerCRUD.create_loader(client)

        with pytest.raises(ToolkitRequiredValueError):
            loader.load_resource_file(trigger_file, {})
        client.config.is_strict_validation = False
        local_dict = loader.load_resource_file(trigger_file, {})[0]
        credentials = loader._authentication_by_id[loader.get_id(local_dict)]
        assert credentials.client_id == "toolkit-client-id"
        assert credentials.client_secret == "toolkit-client-secret"

    def test_credentials_unchanged_changed(self) -> None:
        local_content = """externalId: daily-8am-utc
triggerRule:
  triggerType: schedule
  cronExpression: 0 8 * * *
workflowExternalId: wf_example_repeater
workflowVersion: v1
authentication:
  clientId: my-client-id
  clientSecret: my-client-secret
"""

        cdf_trigger = WorkflowTrigger(
            "daily-8am-utc",
            trigger_rule=WorkflowScheduledTriggerRule(cron_expression="0 8 * * *"),
            workflow_external_id="wf_example_repeater",
            workflow_version="v1",
            metadata={
                WorkflowTriggerCRUD._MetadataKey.secret_hash: calculate_secure_hash(
                    {"clientId": "my-client-id", "clientSecret": "my-client-secret"}, shorten=True
                )
            },
        )
        with monkeypatch_toolkit_client() as client:
            loader = WorkflowTriggerCRUD(client, None, None)

        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = local_content
        local_dumped = loader.load_resource_file(filepath, {})[0]
        cdf_dumped = loader.dump_resource(cdf_trigger, local_dumped)

        assert cdf_dumped == local_dumped

        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = local_content.replace("my-client-secret", "my-client-secret-changed")
        local_dumped = loader.load_resource_file(filepath, {})[0]
        cdf_dumped = loader.dump_resource(cdf_trigger, local_dumped)

        assert cdf_dumped != local_dumped


class TestWorkflowVersionLoader:
    def test_topological_sort_raises_on_cycle(self) -> None:
        dependencies = {
            "a": "b",
            "b": "c",
            "c": "a",  # This creates a cycle
        }

        workflows = [
            WorkflowVersionUpsert(
                workflow_external_id=id_,
                version="v1",
                workflow_definition=WorkflowDefinitionUpsert(
                    tasks=[
                        WorkflowTask(
                            external_id=f"task_{id_}",
                            parameters=SubworkflowReferenceParameters(dependency, "v1"),
                        ),
                    ]
                ),
            )
            for id_, dependency in dependencies.items()
        ]

        with pytest.raises(ToolkitCycleError) as exc:
            WorkflowVersionCRUD.topological_sort(workflows)

        error = exc.value
        assert isinstance(error, ToolkitCycleError)
        assert error.args[1] == [WorkflowVersionId(id_, "v1") for id_ in ["a", "c", "b", "a"]]



from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Sequence
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from functools import cached_property, lru_cache, partial
from typing import ClassVar, Literal, TypeVar, get_args, overload

import questionary
from cognite.client.data_classes import (
    Asset,
    DataSet,
    UserProfileList,
    filters,
)
from cognite.client.data_classes.aggregations import Count
from cognite.client.data_classes.capabilities import (
    UserProfilesAcl,
)
from cognite.client.data_classes.data_modeling import ContainerId, NodeList, Space, SpaceList, View, ViewId, ViewList
from cognite.client.data_classes.data_modeling.statistics import SpaceStatistics
from cognite.client.utils import ms_to_datetime
from questionary import Choice
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.canvas import Canvas
from cognite_toolkit._cdf_tk.client.data_classes.charts import Chart, ChartList, Visibility
from cognite_toolkit._cdf_tk.client.data_classes.migration import ResourceViewMapping
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawTable
from cognite_toolkit._cdf_tk.exceptions import AuthorizationError, ToolkitMissingResourceError, ToolkitValueError

from . import humanize_collection
from .aggregators import (
    AssetAggregator,
    AssetCentricAggregator,
    EventAggregator,
    FileAggregator,
    TimeSeriesAggregator,
)
from .useful_types import AssetCentricDestinationType

T_Type = TypeVar("T_Type", bound=Asset | DataSet)


class AssetCentricInteractiveSelect(ABC):
    def __init__(self, client: ToolkitClient, operation: str) -> None:
        self.client = client
        self.operation = operation
        self._aggregator = self._get_aggregator(client)

    @abstractmethod
    def _get_aggregator(self, client: ToolkitClient) -> AssetCentricAggregator:
        raise NotImplementedError()

    @lru_cache
    def aggregate_count(self, hierarchies: tuple[str, ...], data_sets: tuple[str, ...]) -> int:
        return self._aggregate_count(list(hierarchies), list(data_sets))

    def _aggregate_count(self, hierarchies: list[str], data_sets: list[str]) -> int:
        return self._aggregator.count(hierarchies, data_sets)

    @lru_cache
    def _get_available_data_sets(self, hierarchy: str | None = None) -> list[DataSet]:
        if hierarchy is not None:
            datasets = self._aggregator.used_data_sets(hierarchy)
            return list(self.client.data_sets.retrieve_multiple(external_ids=datasets))
        else:
            return list(self.client.data_sets.list(limit=-1))

    @lru_cache
    def _get_available_hierarchies(self, data_set: str | None = None) -> list[Asset]:
        data_set_external_ids = [data_set] if data_set else None
        return list(self.client.assets.list(root=True, limit=-1, data_set_external_ids=data_set_external_ids))

    def _create_choice(self, item: Asset | DataSet) -> tuple[questionary.Choice, int]:
        """Create a questionary choice for the given item."""
        if item.external_id is None:
            item_count = -1  # No count available for DataSet/Assets without external_id
        elif isinstance(item, DataSet):
            item_count = self.aggregate_count(tuple(), (item.external_id,))
        elif isinstance(item, Asset):
            item_count = self.aggregate_count((item.external_id,), tuple())
        else:
            raise ToolkitValueError(f"Unexpected item type: {type(item)}. Expected Asset or DataSet.")

        return questionary.Choice(
            title=f"{item.name} ({item.external_id}) [{item_count:,}]"
            if item.name != item.external_id
            else f"({item.external_id}) [{item_count:,}]",
            value=item.external_id,
        ), item_count

    @overload
    def select_hierarchy(self, allow_empty: Literal[False] = False) -> str: ...

    @overload
    def select_hierarchy(self, allow_empty: Literal[True]) -> str | None: ...

    def select_hierarchy(self, allow_empty: Literal[False, True] = False) -> str | None:
        """Select a hierarchy interactively."""
        options = self._get_available_hierarchies()
        return self._select("hierarchy", options, allow_empty, "single")

    def select_hierarchies(self) -> list[str]:
        options = self._get_available_hierarchies()
        return self._select("hierarchies", options, False, "multiple")

    @overload
    def select_data_set(self, allow_empty: Literal[False] = False) -> str: ...

    @overload
    def select_data_set(self, allow_empty: Literal[True]) -> str | None: ...

    def select_data_set(self, allow_empty: Literal[False, True] = False) -> str | None:
        """Select a data set interactively."""
        options = self._get_available_data_sets()
        return self._select("data set", options, allow_empty, "single")

    def select_data_sets(self) -> list[str]:
        """Select multiple data sets interactively."""
        options = self._get_available_data_sets()
        return self._select("data sets", options, False, "multiple")

    def select_hierarchies_and_data_sets(self) -> tuple[list[str], list[str]]:
        what = self.select_hierarchies_or_data_sets()
        if what == "Hierarchy":
            hierarchy = self._select("hierarchy", self._get_available_hierarchies(), False, "single")
            data_sets = self._get_available_data_sets(hierarchy=hierarchy)
            if not data_sets:
                return [hierarchy], []
            selected_data_sets = self._select(f"data sets in hierarchy {hierarchy!r}", data_sets, True, "multiple")
            return [hierarchy], selected_data_sets or []
        elif what == "Data Set":
            data_set = self._select("data Set", self._get_available_data_sets(), False, "single")
            hierarchies = self._get_available_hierarchies(data_set=data_set)
            if not hierarchies:
                return [], [data_set]
            selected_hierarchies = self._select(f"hierarchies in data set {data_set!r} ", hierarchies, True, "multiple")
            return selected_hierarchies or [], [data_set]

    def select_hierarchies_or_data_sets(self) -> Literal["Hierarchy", "Data Set"]:
        what = questionary.select(
            f"Do you want to {self.operation} a hierarchy or a data set?", choices=["Hierarchy", "Data Set"]
        ).ask()
        if what is None:
            raise ToolkitValueError("No selection made. Aborting.")
        if what not in ["Hierarchy", "Data Set"]:
            raise ToolkitValueError(f"Unexpected selection: {what}. Aborting.")
        return what

    @overload
    def _select(
        self,
        what: str,
        options: Sequence[T_Type],
        allow_empty: Literal[False],
        plurality: Literal["single"],
    ) -> str: ...

    @overload
    def _select(
        self,
        what: str,
        options: Sequence[T_Type],
        allow_empty: Literal[True],
        plurality: Literal["single"],
    ) -> str | None: ...

    @overload
    def _select(
        self,
        what: str,
        options: Sequence[T_Type],
        allow_empty: Literal[False],
        plurality: Literal["multiple"],
    ) -> list[str]: ...

    @overload
    def _select(
        self,
        what: str,
        options: Sequence[T_Type],
        allow_empty: Literal[True],
        plurality: Literal["multiple"],
    ) -> list[str] | None: ...

    def _select(
        self,
        what: str,
        options: Sequence[Asset | DataSet],
        allow_empty: Literal[True, False],
        plurality: Literal["single", "multiple"],
    ) -> str | list[str] | None:
        """Select a single item interactively."""
        if not options and not allow_empty:
            raise ToolkitValueError(f"No {what} available to select.")

        choices: list[questionary.Choice] = []
        none_sentinel = object()
        if allow_empty:
            choices.append(questionary.Choice(title=f"All {what}", value=none_sentinel))

        for choice, count in sorted(
            (self._create_choice(item) for item in options), key=lambda x: (-x[1], x[0].title)
        ):  # cont, choice.title
            if count > 0:
                choices.append(choice)
        if not choices and not allow_empty:
            raise ToolkitValueError(f"No {what} available with data to select.")

        message = f"Select a {what} to {self.operation} listed as 'name (external_id) [count]'"
        if plurality == "multiple":
            selected = questionary.checkbox(message, choices=choices).ask()
        else:
            selected = questionary.select(message, choices=choices).ask()
        if selected is None:
            raise ToolkitValueError(f"No {what} selected. Aborting.")
        elif selected is none_sentinel or (isinstance(selected, list) and none_sentinel in selected):
            return None
        return selected


class AssetInteractiveSelect(AssetCentricInteractiveSelect):
    def _get_aggregator(self, client: ToolkitClient) -> AssetCentricAggregator:
        return AssetAggregator(self.client)


class FileMetadataInteractiveSelect(AssetCentricInteractiveSelect):
    def _get_aggregator(self, client: ToolkitClient) -> AssetCentricAggregator:
        return FileAggregator(self.client)


class TimeSeriesInteractiveSelect(AssetCentricInteractiveSelect):
    def _get_aggregator(self, client: ToolkitClient) -> AssetCentricAggregator:
        return TimeSeriesAggregator(self.client)


class EventInteractiveSelect(AssetCentricInteractiveSelect):
    def _get_aggregator(self, client: ToolkitClient) -> AssetCentricAggregator:
        return EventAggregator(self.client)


class RawTableInteractiveSelect:
    def __init__(self, client: ToolkitClient, operation: str) -> None:
        self.client = client
        self.operation = operation

    def _available_databases(self) -> list[str]:
        databases = self.client.raw.databases.list(limit=-1)
        return [db.name for db in databases if db.name is not None]

    def _available_tables(self, database: str) -> list[RawTable]:
        tables = self.client.raw.tables.list(db_name=database, limit=-1)
        return [RawTable(database, table.name) for table in tables if table.name is not None]

    def select_tables(self, database: str | None = None) -> list[RawTable]:
        """Interactively select raw tables."""
        databases = self._available_databases()
        if not databases:
            raise ToolkitValueError("No raw databases available. Aborting.")
        if database and database not in databases:
            raise ToolkitValueError(
                f"Database '{database}' not found in available raw databases: {databases}. Aborting."
            )
        elif database:
            selected_database = database
        else:
            selected_database = questionary.select(
                f"Select a Raw Database to {self.operation}",
                choices=[questionary.Choice(title=db, value=db) for db in databases],
            ).ask()
        if selected_database is None:
            raise ToolkitValueError("No database selected. Aborting.")
        available_tables = self._available_tables(selected_database)
        if not available_tables:
            raise ToolkitValueError(f"No raw tables available in database '{selected_database}'. Aborting.")

        selected_tables = questionary.checkbox(
            f"Select Raw Tables in {selected_database} to {self.operation}",
            choices=[questionary.Choice(title=f"{table.table_name}", value=table) for table in available_tables],
        ).ask()

        if selected_tables is None:
            raise ToolkitValueError("No tables selected. Aborting.")
        return selected_tables


@dataclass
class CanvasFilter:
    visibility: Literal["public", "private"] | None = None
    created_by: Literal["user"] | None = None
    select_all: bool = False

    def as_dms_filter(self) -> filters.Filter:
        canvas_id = Canvas.get_source()
        leaf_filters: list[filters.Filter] = [
            filters.Not(filters.Equals(canvas_id.as_property_ref("isArchived"), True)),
            # When sourceCanvasId is not set, we get the newest version of the canvas
            filters.Not(filters.Exists(canvas_id.as_property_ref("sourceCanvasId"))),
        ]
        if self.visibility is not None:
            leaf_filters.append(filters.Equals(canvas_id.as_property_ref("visibility"), self.visibility))

        return filters.And(*leaf_filters)


class InteractiveCanvasSelect:
    opening_choices: ClassVar[list[questionary.Choice]] = [
        questionary.Choice(title="All public Canvases", value=CanvasFilter(visibility="public", select_all=True)),
        questionary.Choice(title="Selected public Canvases", value=CanvasFilter(visibility="public", select_all=False)),
        questionary.Choice(title="All by given user", value=CanvasFilter(created_by="user", select_all=True)),
        questionary.Choice(title="Selected by given user", value=CanvasFilter(created_by="user", select_all=False)),
        questionary.Choice(title="All Canvases", value=CanvasFilter(visibility=None, select_all=True)),
    ]

    def __init__(self, client: ToolkitClient) -> None:
        self.client = client

    def select_external_ids(self) -> list[str]:
        select_filter = self._select_filter()

        return self._select_external_ids(select_filter)

    @classmethod
    def _select_filter(cls) -> CanvasFilter:
        user_response = questionary.select(
            "Which Canvases do you want to select?",
            choices=cls.opening_choices,
        ).ask()
        if user_response is None:
            raise ToolkitValueError("No Canvas selection made. Aborting.")
        return user_response

    def _select_external_ids(self, select_filter: CanvasFilter) -> list[str]:
        available_canvases = self.client.canvas.list(filter=select_filter.as_dms_filter(), limit=-1)
        if select_filter.select_all and select_filter.created_by is None:
            return [canvas.external_id for canvas in available_canvases]
        users = self.client.iam.user_profiles.list(limit=-1)
        display_name_by_user_identifier = {user.user_identifier: user.display_name or "missing" for user in users}
        if select_filter.created_by == "user":
            canvas_by_user: dict[str, list[Canvas]] = defaultdict(list)
            for canvas in available_canvases:
                canvas_by_user[canvas.created_by].append(canvas)

            user_response = questionary.select(
                "Which user do you want to select Canvases for?",
                choices=[
                    questionary.Choice(
                        title=f"{user.display_name} ({user.user_identifier})",
                        value=canvas_by_user[user.user_identifier],
                    )
                    for user in users
                    if user.user_identifier in canvas_by_user
                ],
            ).ask()
            available_canvases = NodeList[Canvas](user_response)

        if select_filter.select_all:
            return [canvas.external_id for canvas in available_canvases]

        selected_canvases = questionary.checkbox(
            "Select Canvases",
            choices=[
                questionary.Choice(
                    title=f"{canvas.name} (Created by {display_name_by_user_identifier[canvas.created_by]!r}, last updated {canvas.updated_at!r})",
                    value=canvas.external_id,
                )
                for canvas in available_canvases
            ],
        ).ask()

        return selected_canvases or []


@dataclass
class ChartFilter:
    visibility: Visibility | None = None
    owned_by: Literal["user"] | None = None
    select_all: bool = False


class InteractiveChartSelect:
    opening_choices: ClassVar[list[questionary.Choice]] = [
        questionary.Choice(title="All public Charts", value=ChartFilter(visibility="PUBLIC", select_all=True)),
        questionary.Choice(title="Selected public Charts", value=ChartFilter(visibility="PUBLIC", select_all=False)),
        questionary.Choice(title="All owned by given user", value=ChartFilter(owned_by="user", select_all=True)),
        questionary.Choice(title="Selected owned by given user", value=ChartFilter(owned_by="user", select_all=False)),
    ]

    def __init__(self, client: ToolkitClient) -> None:
        self.client = client

    def select_external_ids(self) -> list[str]:
        select_filter = self._select_filter()
        return self._select_external_ids(select_filter)

    @classmethod
    def _select_filter(cls) -> ChartFilter:
        user_response = questionary.select(
            "Which Charts do you want to select?",
            choices=cls.opening_choices,
        ).ask()
        if user_response is None:
            raise ToolkitValueError("No Chart selection made. Aborting.")
        return user_response

    def _select_external_ids(self, select_filter: ChartFilter) -> list[str]:
        if not self.client.iam.verify_capabilities(
            UserProfilesAcl([UserProfilesAcl.Action.Read], scope=UserProfilesAcl.Scope.All())
        ):
            raise AuthorizationError(
                "The current user does not have permission to list user profiles, "
                "which is required to select Charts owned by a specific user."
            )
        available_charts = self.client.charts.list(visibility=(select_filter.visibility or "PUBLIC"))
        if select_filter.select_all and select_filter.owned_by is None:
            return [chart.external_id for chart in available_charts]

        users = self.client.iam.user_profiles.list(limit=-1)
        display_name_by_user_identifier = {
            user.user_identifier: user.display_name for user in users if user.display_name
        }
        if select_filter.owned_by == "user":
            available_charts = self._select_charts_by_user(available_charts, users)
        if select_filter.select_all:
            return [chart.external_id for chart in available_charts]
        selected_charts = questionary.checkbox(
            "Select Charts",
            choices=[
                questionary.Choice(
                    title=f"{chart.data.name} (Created by "
                    f"{display_name_by_user_identifier.get(chart.owner_id, 'missing')!r} "
                    f"- {ms_to_datetime(chart.last_updated_time)})",
                    value=chart.external_id,
                )
                for chart in available_charts
            ],
        ).ask()
        return selected_charts or []

    @classmethod
    def _select_charts_by_user(cls, available_charts: ChartList, users: UserProfileList) -> ChartList:
        chart_by_user: dict[str, list[Chart]] = defaultdict(list)
        for chart in available_charts:
            chart_by_user[chart.owner_id].append(chart)
        user_response = questionary.select(
            "Which user do you want to select Charts for?",
            choices=[
                questionary.Choice(
                    title=f"{user.display_name} ({user.user_identifier})",
                    value=chart_by_user[user.user_identifier],
                )
                for user in users
                if user.user_identifier in chart_by_user
            ],
        ).ask()
        available_charts = ChartList(user_response)
        return available_charts


class AssetCentricDestinationSelect:
    valid_destinations: tuple[str, ...] = get_args(AssetCentricDestinationType)

    @classmethod
    def validate(cls, destination_type: str) -> AssetCentricDestinationType:
        if destination_type not in cls.valid_destinations:
            raise ToolkitValueError(
                f"Invalid destination type: {destination_type!r}. Must be one of {humanize_collection(cls.valid_destinations)}."
            )
        # We validated the destination type above
        return destination_type  # type: ignore[return-value]

    @classmethod
    def select(cls) -> AssetCentricDestinationType:
        """Interactively select a destination type."""
        destination_type = questionary.select(
            "Select a destination type",
            choices=[questionary.Choice(title=dest.capitalize(), value=dest) for dest in cls.valid_destinations],
        ).ask()
        if destination_type is None:
            raise ToolkitValueError("No destination type selected. Aborting.")
        # We only input valid destination types, so we can safely skip MyPy's type checking here
        return destination_type

    @classmethod
    def get(cls, destination_type: str | None = None) -> AssetCentricDestinationType:
        """Get the destination type, either from the input or interactively."""
        if destination_type is None:
            return cls.select()
        return cls.validate(destination_type)


class DataModelingSelect:
    """A utility class to select Data Modeling nodes interactively."""

    def __init__(self, client: ToolkitClient, operation: str, console: Console | None = None) -> None:
        self.client = client
        self.operation = operation
        self.console = console or Console()
        self._available_spaces: SpaceList | None = None

    @cached_property
    def stats_by_space(self) -> dict[str, SpaceStatistics]:
        result = self.client.data_modeling.statistics.spaces.list()
        return {stat.space: stat for stat in result}

    @overload
    def select_view(
        self,
        multiselect: Literal[False] = False,
        include_global: bool = False,
        space: str | None = None,
        message: str | None = None,
        instance_type: Literal["node", "edge", "all"] | None = None,
        mapped_container: ContainerId | None = None,
    ) -> View: ...

    @overload
    def select_view(
        self,
        multiselect: Literal[True],
        include_global: bool = False,
        space: str | None = None,
        message: str | None = None,
        instance_type: Literal["node", "edge", "all"] | None = None,
        mapped_container: ContainerId | None = None,
    ) -> ViewList: ...

    def select_view(
        self,
        multiselect: bool = False,
        include_global: bool = False,
        space: str | None = None,
        message: str | None = None,
        instance_type: Literal["node", "edge", "all"] | None = None,
        mapped_container: ContainerId | None = None,
    ) -> View | ViewList:
        """Select one or more views interactively.

        Args:
            multiselect: Whether to allow selecting multiple views.
            include_global: Whether to include global views in the selection.
            space: The space to select views from. If None, the user will be prompted to
                select a space.
            message: The message to display when prompting for a view. If None, a default message
                will be used.
            instance_type: If 'node' or 'edge', only views of that type will be shown.
            mapped_container: Only selects view(s) that are mapping the given container.

        Returns:
            The selected view(s).

        Raises:
            ToolkitValueError: If no view(s) are selected or if the selected view(s) are not valid.
            ToolkitMissingResourceError: If no views are found in the selected space.

        """
        selected_space = (
            space
            or self.select_schema_space(
                include_global,
                message=f"In which Spaces is the view you will use to select instances to {self.operation}?",
            ).space
        )

        views = self.client.data_modeling.views.list(
            space=selected_space,
            include_inherited_properties=True,
            limit=-1,
            include_global=include_global,
        )
        views = ViewList(
            [
                view
                for view in views
                if instance_type in (None, "all", view.used_for)
                and (mapped_container is None or mapped_container in view.referenced_containers())
            ]
        )
        if not views:
            raise ToolkitMissingResourceError(f"No views found in space {selected_space!r}.")
        question = message or f"Which view do you want to use to select instances to {self.operation}?"
        choices = [Choice(title=f"{view.external_id} (version={view.version})", value=view) for view in views]
        if multiselect:
            selected_views = questionary.checkbox(question, choices=choices).ask()
        else:
            selected_views = questionary.select(question, choices=choices).ask()
        if selected_views is None:
            raise ToolkitValueError("No view(s) selected")
        if multiselect:
            if not isinstance(selected_views, list) or not all(isinstance(v, View) for v in selected_views):
                raise ToolkitValueError(f"Selected views is not a valid list of View objects: {selected_views!r}")
            return ViewList(selected_views)
        else:
            if not isinstance(selected_views, View):
                raise ToolkitValueError(f"Selected view is not a valid View object: {selected_views!r}")
            return selected_views

    def select_schema_space(self, include_global: bool, message: str | None = None) -> Space:
        message = message or f"Select the space to {self.operation}:"
        spaces = self._get_available_spaces(include_global)
        schema_spaces = {
            space
            for space, stats in self.stats_by_space.items()
            if (stats.containers + stats.views + stats.data_models) > 0
        }
        spaces = SpaceList([space for space in spaces if space.space in schema_spaces or space.is_global])
        if not spaces:
            raise ToolkitMissingResourceError("No spaces with schema (containers, views, or data models) found.")
        selected_space = questionary.select(
            message,
            [Choice(title=space.space, value=space) for space in sorted(spaces, key=lambda s: s.space)],
        ).ask()
        if selected_space is None:
            raise ToolkitValueError("No space selected")
        if not isinstance(selected_space, Space):
            raise ToolkitValueError(f"Selected space is not a valid Space object: {selected_space!r}")
        return selected_space

    def select_instance_type(
        self, view_used_for: Literal["node", "edge", "all"] | None = None
    ) -> Literal["node", "edge"]:
        """Selects an instance type (node or edge) interactively.

        Args:
            view_used_for: If 'node' or 'edge', that type is returned directly.
                           If 'all' or None, the user is prompted to select.

        Returns:
            The selected instance type, either 'node' or 'edge'.
        """
        if view_used_for is not None and view_used_for != "all":
            return view_used_for
        selected_instance_type = questionary.select(
            f"What type of instances do you want to {self.operation}?",
            choices=[
                Choice(title="Nodes", value="node"),
                Choice(title="Edges", value="edge"),
            ],
        ).ask()
        if selected_instance_type is None:
            raise ToolkitValueError("No instance type selected")
        if selected_instance_type not in ("node", "edge"):
            raise ToolkitValueError(f"Selected instance type is not valid: {selected_instance_type!r}")
        return selected_instance_type

    def select_space_type(self) -> Literal["instance", "schema", "empty"]:
        selected_space_type = questionary.select(
            f"What type of spaces do you want to {self.operation}?",
            choices=[
                Choice(title="Instance spaces with instances", value="instance"),
                Choice(title="Schema spaces with schema", value="schema"),
                Choice(title="Empty spaces without schema or instances", value="empty"),
            ],
        ).ask()
        if selected_space_type is None:
            raise ToolkitValueError("No space type selected")
        if selected_space_type not in ["instance", "schema", "empty"]:
            raise ToolkitValueError(f"Selected space type is not valid: {selected_space_type!r}")
        return selected_space_type

    @overload
    def select_instance_space(
        self,
        multiselect: Literal[False],
        selected_view: ViewId | None = None,
        instance_type: Literal["node", "edge"] | None = None,
        message: str | None = None,
        include_empty: bool = False,
    ) -> str: ...

    @overload
    def select_instance_space(
        self,
        multiselect: Literal[True] = True,
        selected_view: ViewId | None = None,
        instance_type: Literal["node", "edge"] | None = None,
        message: str | None = None,
        include_empty: bool = False,
    ) -> list[str]: ...

    def select_instance_space(
        self,
        multiselect: bool = True,
        selected_view: ViewId | None = None,
        instance_type: Literal["node", "edge"] | None = None,
        message: str | None = None,
        include_empty: bool = False,
    ) -> str | list[str]:
        if (selected_view is not None and instance_type is None) or (
            selected_view is None and instance_type is not None
        ):
            raise ToolkitValueError("Both selected_view and instance_type must be provided or both must be None.")
        if selected_view is not None and instance_type is not None:
            all_spaces = self._get_available_spaces(include_global=False)
            count_by_space = self._get_instance_count_in_view_by_space(all_spaces, selected_view, instance_type)
        else:
            count_by_space = {
                space: stats.nodes + stats.edges
                for space, stats in self.stats_by_space.items()
                if (stats.nodes + stats.edges) > 0
                or (
                    include_empty
                    and (stats.nodes + stats.edges + stats.containers + stats.views + stats.data_models) == 0
                )
            }

        if not count_by_space:
            raise ToolkitMissingResourceError(
                f"No instances found in any space for the view {selected_view!r} with instance type {instance_type!r}."
            )
        if len(count_by_space) == 1:
            selected_spaces = next(iter(count_by_space.keys()))
            self.console.print(f"Only one space with instances found: {selected_spaces!r}. Using this space.")
            return [selected_spaces] if multiselect else selected_spaces

        if not message:
            message = f"In which Space{'(s)' if multiselect else ''} do you want to {self.operation} instances?"
        choices = [
            Choice(title=f"{space} ({count:,} instances)", value=space)
            for space, count in sorted(count_by_space.items(), key=lambda item: item[1], reverse=True)
        ]
        if multiselect:
            selected_spaces = questionary.checkbox(message, choices=choices).ask()
        else:
            selected_spaces = questionary.select(message, choices=choices).ask()
        if selected_spaces is None or (isinstance(selected_spaces, list) and len(selected_spaces) == 0):
            raise ToolkitValueError(
                f"No instance space selected to {self.operation}. Use arrow keys to navigate and space key to select. Press enter to confirm."
            )
        return selected_spaces

    @overload
    def select_empty_spaces(self, multiselect: Literal[False]) -> str: ...
    @overload
    def select_empty_spaces(self, multiselect: Literal[True]) -> list[str]: ...

    def select_empty_spaces(self, multiselect: bool = True) -> str | list[str]:
        empty_spaces = [
            space
            for space, stats in self.stats_by_space.items()
            if (stats.nodes + stats.edges + stats.containers + stats.views + stats.data_models) == 0
        ]
        if not empty_spaces:
            raise ToolkitMissingResourceError("No empty spaces found.")
        if len(empty_spaces) == 1:
            selected_space = empty_spaces[0]
            self.console.print(f"Only one empty space found: {selected_space!r}. Using this space.")
            return [selected_space] if multiselect else selected_space

        message = f"In which empty Space{'(s)' if multiselect else ''} do you want to {self.operation}?"
        choices = [Choice(title=f"{space}", value=space) for space in sorted(empty_spaces)]
        if multiselect:
            selected_spaces = questionary.checkbox(message, choices=choices).ask()
        else:
            selected_spaces = questionary.select(message, choices=choices).ask()
        if selected_spaces is None or (isinstance(selected_spaces, list) and len(selected_spaces) == 0):
            raise ToolkitValueError(f"No empty space selected to {self.operation}. Aborting.")
        return selected_spaces

    def _get_instance_count_in_view_by_space(
        self, all_spaces: SpaceList, view_id: ViewId, instance_type: Literal["node", "edge"]
    ) -> dict[str, float]:
        count_by_space: dict[str, float] = {}
        result = self.client.data_modeling.statistics.project()
        read_limit = result.concurrent_read_limit

        with ThreadPoolExecutor(max_workers=read_limit // 2) as executor:
            results = executor.map(
                partial(self._instance_count_space, view_id=view_id, instance_type=instance_type),
                (space.space for space in all_spaces),
            )
            for space_name, count in results:
                if count > 0:
                    count_by_space[space_name] = count

        return count_by_space

    @lru_cache
    def _instance_count_space(
        self, space: str, view_id: ViewId, instance_type: Literal["node", "edge"]
    ) -> tuple[str, float]:
        """Get the count of instances in a specific space for a given view and instance type."""
        return space, self.client.data_modeling.instances.aggregate(
            view_id, Count("externalId"), instance_type=instance_type, space=space
        ).value or 0.0

    def _get_available_spaces(self, include_global: bool = False) -> SpaceList:
        if self._available_spaces is None:
            self._available_spaces = self.client.data_modeling.spaces.list(include_global=True, limit=-1)
        if include_global:
            return self._available_spaces
        return SpaceList([space for space in self._available_spaces if not space.is_global])


class ResourceViewMappingInteractiveSelect:
    def __init__(self, client: ToolkitClient, operation: str) -> None:
        self.client = client
        self.operation = operation

    def select_resource_view_mapping(self, resource_type: str) -> ResourceViewMapping:
        """Select a Resource View Mapping interactively.

        Args:
            resource_type: The resource type to filter Resource View Mappings by.
        Returns:
            The selected Resource View Mapping.
        """
        mappings = self.client.migration.resource_view_mapping.list(resource_type=resource_type, limit=-1)
        if not mappings:
            raise ToolkitMissingResourceError(f"No Resource View Mappings found for resource type {resource_type!r}.")
        choices = [
            Choice(
                title=f"{mapping.external_id} ({mapping.view_id!r})",
                value=mapping,
            )
            for mapping in mappings
        ]
        selected_mapping = questionary.select(
            f"Which Resource View Mapping do you want to use to {self.operation}? [identifier] (ingestion view)",
            choices=choices,
        ).ask()
        if selected_mapping is None:
            raise ToolkitValueError("No Resource View Mapping selected.")
        if not isinstance(selected_mapping, ResourceViewMapping):
            raise ToolkitValueError(
                f"Selected Resource View Mapping is not a valid ResourceViewMapping object: {selected_mapping!r}"
            )
        return selected_mapping



from dataclasses import dataclass
from itertools import groupby
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.cruds import ResourceTypes
from cognite_toolkit._cdf_tk.data_classes import (
    BuildVariables,
    BuiltModule,
    BuiltModuleList,
    BuiltResource,
    BuiltResourceList,
    SourceLocationEager,
)


@dataclass
class GetResourcesArgs:
    resource_dir: ResourceTypes
    kind: str | None
    selected: str | Path | None


class TestBuiltModuleList:
    # Anchor for absolute paths in tests
    anchor = f"{Path.cwd()}/"

    @pytest.mark.parametrize(
        "module,args,expected",
        [
            pytest.param(
                {
                    Path(f"{anchor}modules/module1"): [
                        Path(f"{anchor}modules/module1/transformations/my.Transformation.yaml"),
                        Path(f"{anchor}modules/module1/transformations/my.Schedule.yaml"),
                    ]
                },
                GetResourcesArgs(
                    resource_dir="transformations",
                    kind="Transformation",
                    selected="module1",
                ),
                [Path(f"{anchor}modules/module1/transformations/my.Transformation.yaml")],
                id="Select by module name",
            ),
            pytest.param(
                {
                    Path(f"{anchor}modules/module1"): [
                        Path(f"{anchor}modules/module1/transformations/my.Transformation.yaml"),
                        Path(f"{anchor}modules/module1/transformations/my.Schedule.yaml"),
                    ]
                },
                GetResourcesArgs(
                    resource_dir="transformations",
                    kind="Transformation",
                    selected=Path("modules/module1"),
                ),
                [Path(f"{anchor}modules/module1/transformations/my.Transformation.yaml")],
                id="Select with relative with module in absolute",
            ),
            pytest.param(
                {
                    Path("modules/module1"): [
                        Path("modules/module1/transformations/my.Transformation.yaml"),
                        Path("modules/module1/transformations/my.Schedule.yaml"),
                    ]
                },
                GetResourcesArgs(
                    resource_dir="transformations",
                    kind="Transformation",
                    selected=Path(f"{anchor}modules/module1"),
                ),
                [Path("modules/module1/transformations/my.Transformation.yaml")],
                id="Select with absolute with module in relative",
            ),
            pytest.param(
                {
                    Path("modules/module1"): [
                        Path(f"{anchor}modules/module1/transformations/my.Transformation.yaml"),
                        Path(f"{anchor}modules/module1/transformations/my.Schedule.yaml"),
                    ]
                },
                GetResourcesArgs(
                    resource_dir="transformations",
                    kind="Transformation",
                    selected=Path(f"{anchor}modules/module1/transformations/my.Transformation.yaml"),
                ),
                [Path(f"{anchor}modules/module1/transformations/my.Transformation.yaml")],
                id="Select file by absolute path",
            ),
            pytest.param(
                {
                    Path("module1"): [
                        Path("modules/module1/transformations/my.Transformation.yaml"),
                        Path("modules/module1/transformations/my.Schedule.yaml"),
                    ]
                },
                GetResourcesArgs(
                    resource_dir="transformations",
                    kind="Transformation",
                    selected=Path("modules/module1/transformations/my.Transformation.yaml"),
                ),
                [Path("modules/module1/transformations/my.Transformation.yaml")],
                id="Select file by relative path",
            ),
            pytest.param(
                {
                    Path("modules/module1"): [
                        Path(f"{anchor}modules/module1/transformations/my.Transformation.yaml"),
                        Path(f"{anchor}modules/module1/transformations/my.Schedule.yaml"),
                    ]
                },
                GetResourcesArgs(
                    resource_dir="transformations",
                    kind="Transformation",
                    selected=Path(f"{anchor}modules/module1/transformations"),
                ),
                [Path(f"{anchor}modules/module1/transformations/my.Transformation.yaml")],
                id="Select resource dir by absolute path",
            ),
            pytest.param(
                {
                    Path("module1"): [
                        Path("modules/module1/transformations/my.Transformation.yaml"),
                        Path("modules/module1/transformations/my.Schedule.yaml"),
                    ]
                },
                GetResourcesArgs(
                    resource_dir="transformations",
                    kind="Transformation",
                    selected=Path("modules/module1/transformations/my.Transformation.yaml"),
                ),
                [Path("modules/module1/transformations/my.Transformation.yaml")],
                id="Select resource dir by relative path",
            ),
            pytest.param(
                {
                    Path(f"{anchor}module1"): [
                        Path(f"{anchor}modules/module1/transformations/my.Transformation.yaml"),
                        Path(f"{anchor}modules/module1/transformations/my.Schedule.yaml"),
                    ],
                    Path(f"{anchor}module2"): [
                        Path(f"{anchor}module2/transformations/other.Transformation.yaml"),
                    ],
                },
                GetResourcesArgs(
                    resource_dir="transformations",
                    kind="Transformation",
                    selected=None,
                ),
                [
                    Path(f"{anchor}modules/module1/transformations/my.Transformation.yaml"),
                    Path(f"{anchor}module2/transformations/other.Transformation.yaml"),
                ],
                id="Select all by kind",
            ),
            pytest.param(
                {
                    Path(f"{anchor}module1"): [
                        Path(f"{anchor}modules/module1/transformations/my.Transformation.yaml"),
                    ]
                },
                GetResourcesArgs(
                    resource_dir="transformations",
                    kind="NonExistentKind",
                    selected=None,
                ),
                [],
                id="Select with non-existent kind",
            ),
            pytest.param(
                {
                    Path("modules/module1"): [
                        Path(f"{anchor}modules/module1/transformations/my.Transformation.yaml"),
                    ]
                },
                GetResourcesArgs(
                    resource_dir="transformations",
                    kind="Transformation",
                    selected=Path("/modules/module1/transformations/doesnotexist.yaml"),
                ),
                [],
                id="Select with non-existent path",
            ),
            pytest.param(
                {
                    Path("modules/module1"): [
                        Path(f"{anchor}modules/module1/transformations/my.Transformation.yaml"),
                    ]
                },
                GetResourcesArgs(
                    resource_dir="transformations",
                    kind="Transformation",
                    selected="notamodule",
                ),
                [],
                id="Select with non-existent module name",
            ),
        ],
    )
    def test_get_resources_selected(
        self,
        module: dict[Path, list[Path]],
        args: GetResourcesArgs,
        expected: list[Path],
    ) -> None:
        module_list = self._create_built_resource_list(module)
        result = module_list.get_resources(
            id_type=None,
            resource_dir=args.resource_dir,
            kind=args.kind,
            selected=args.selected,
        )
        actual = [item.source.path for item in result]
        assert actual == expected

    @staticmethod
    def _create_built_resource_list(module: dict[Path, list[Path]]) -> BuiltModuleList:
        modules: list[BuiltModule] = [
            BuiltModule(
                name=module_path.name,
                location=SourceLocationEager(module_path, "hash1234"),
                build_variables=BuildVariables([]),
                resources={
                    resource_dir: BuiltResourceList(
                        [
                            BuiltResource(
                                identifier=f"resource_{i}",
                                source=SourceLocationEager(resource_path, "hash5678"),
                                kind=resource_path.stem.split(".")[-1],
                                destination=None,
                                extra_sources=None,
                            )
                            for i, resource_path in enumerate(resource_paths)
                        ]
                    )
                    for resource_dir, resource_paths in groupby(
                        sorted(resource_paths, key=lambda p: p.parent.name), key=lambda p: p.parent.name
                    )
                },
                warning_count=0,
                status="success",
                iteration=1,
            )
            for module_path, resource_paths in module.items()
        ]
        module_list = BuiltModuleList(modules)
        return module_list



import pytest
from cognite.client import global_config

TOKEN_URL = "https://test.com/token"


@pytest.fixture
def max_retries_2():
    old = global_config.max_retries
    global_config.max_retries = 2
    yield
    global_config.max_retries = old
